# -*- coding: utf-8 -*-
"""
biosppy.signals.pcg
-------------------

This module provides methods to process Phonocardiography (PCG) signals.

:copyright: (c) 2015-2018 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# Imports
# compat
from __future__ import absolute_import, division, print_function

# 3rd party
import numpy as np
import scipy.signal as ss
import matplotlib.pyplot as plt
from matplotlib.patches import Patch

# local
from . import tools as st
from . import ecg
from .. import plotting, utils


def pcg(signal=None, sampling_rate=1000., units=None, path=None, show=True):
    """

    Parameters
    ----------
    signal : array
        Raw PCG signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    units : str, optional
        The units of the input signal. If specified, the plot will have the
        y-axis labeled with the corresponding units.
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show a summary plot.

    Returns
    -------
    ts : array
        Signal time axis reference (seconds).
    filtered : array
        Filtered PCG signal.
    peaks : array
        Peak location indices.
    hs: array
        Classification of peaks as S1 or S2.
    heart_rate : double
        Average heart rate (bpm).
    systolic_time_interval : double
        Average systolic time interval (seconds).
    heart_rate_ts : array
         Heart rate time axis reference (seconds).
    inst_heart_rate : array
        Instantaneous heart rate (bpm).

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    sampling_rate = float(sampling_rate)

    # Filter Design
    order = 2
    passBand = np.array([25, 400])
    
    # Band-Pass filtering of the PCG:        
    filtered, fs, params = st.filter_signal(signal, 'butter', 'bandpass', order, passBand, sampling_rate)

    # find peaks
    peaks, envelope = find_peaks(signal=filtered, sampling_rate=sampling_rate)
    
    # classify heart sounds
    hs, = identify_heart_sounds(beats=peaks, sampling_rate=sampling_rate)
    s1_peaks = peaks[np.where(hs==1)[0]]
    
    # get heart rate
    heartRate,systolicTimeInterval = get_avg_heart_rate(envelope,sampling_rate)
    
    # get instantaneous heart rate
    hr_idx,hr = st.get_heart_rate(s1_peaks, sampling_rate)
    
    # get time vectors
    length = len(signal)
    T = (length - 1) / sampling_rate
    ts = np.linspace(0, T, length, endpoint=True)
    ts_hr = ts[hr_idx]

    # plot
    if show:
        plotting.plot_pcg(ts=ts,
                raw=signal,
                filtered=filtered,
                peaks=peaks,
                heart_sounds=hs,
                heart_rate_ts=ts_hr,
                inst_heart_rate=hr,
                units=units,
                path=path,
                show=True)
        
        
    # output
    args = (ts, filtered, peaks, hs, heartRate, systolicTimeInterval, ts_hr, hr)
    names = ('ts', 'filtered', 'peaks', 'heart_sounds',
             'heart_rate', 'systolic_time_interval','heart_rate_ts','inst_heart_rate')

    return utils.ReturnTuple(args, names)

def find_peaks(signal=None,sampling_rate=1000.):
    
    """Finds the peaks of the heart sounds from the homomorphic envelope

    Parameters
    ----------
    signal : array
        Input filtered PCG signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).

    Returns
    -------
    peaks : array
        peak location indices.
    envelope : array
        Homomorphic envelope (normalized).

    """
    
    # Compute homomorphic envelope
    envelope, = homomorphic_filter(signal,sampling_rate)
    envelope, = st.normalize(envelope)
    
    # Find the prominent peaks of the envelope
    peaksIndices, _ = ss.find_peaks(envelope, distance = 0.10*sampling_rate, prominence = 0.25)
    
    peaks = np.array(peaksIndices, dtype='int')

    return utils.ReturnTuple((peaks,envelope), 
                             ('peaks','homomorphic_envelope'))


def homomorphic_filter(signal=None, sampling_rate=1000., f_LPF=8, order=2):
    """Finds the homomorphic envelope of a signal.

    Adapted to Python from original MATLAB code written by David Springer, 2016 (C), for
    comparison purposes in the paper [Springer15]_.
    Available at: https://github.com/davidspringer/Springer-Segmentation-Code

    Follows the approach described by Schmidt et al. [Schimdt10]_.

    Parameters
    ----------
    signal : array
        Input filtered PCG signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    f_LPF: int, float, optional
        Low pass cut-off frequency (Hz)
    order: int, optional
        Order of Butterworth low pass filter.

    Returns
    -------
    envelope : array
        Homomorphic envelope (non-normalized).

    References
    ----------
    .. [Springer15] D.Springer, "Logistic Regression-HSMM-based Heart Sound Segmentation",
       IEEE Trans. Biomed. Eng., In Press, 2015.
    .. [Schimdt10] S. E. Schmidt et al., "Segmentation of heart sound recordings by a 
       duration-dependent hidden Markov model", Physiol. Meas., 2010

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    sampling_rate = float(sampling_rate)
    f_LPF = float(f_LPF)
    
    # Filter Design
    passBand = np.array([25, 400])
    
    # Band-Pass filtering of the PCG:        
    signal, fs, params = st.filter_signal(signal, 'butter', 'bandpass', order, passBand, sampling_rate)
    
    # LP-filter Design (to reject the oscillating component of the signal):
    b, a = ss.butter(order, 2 * f_LPF / fs, 'low')
    envelope = np.exp(ss.filtfilt(b, a, np.log(np.abs(ss.hilbert(signal)))))
    
    # Remove spurious spikes in first sample:
    envelope[0] = envelope[1]   

    return utils.ReturnTuple((envelope,), 
                             ('homomorphic_envelope',))

def get_avg_heart_rate(envelope=None, sampling_rate=1000.):
    
    """Compute average heart rate from the signal's homomorphic envelope.
    
    Follows the approach described by Schmidt et al. [Schimdt10]_, with
    code adapted from David Springer [Springer16]_.
    
    Parameters
    ----------
    envelope : array
        Signal's homomorphic envelope
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
        
    Returns
    -------
    heart_rate : double
        Average heart rate (bpm).
    systolic_time_interval : double
        Average systolic time interval (seconds).

    Notes
    -----
    * Assumes normal human heart rate to be between 40 and 200 bpm.
    * Assumes normal human systole time interval to be between 0.2 seconds and half a heartbeat
    
    References
    ----------
    .. [Schimdt10] S. E. Schmidt et al., "Segmentation of heart sound recordings by a 
       duration-dependent hidden Markov model", Physiol. Meas., 2010
    .. [Springer16] D.Springer, "Heart sound segmentation code based on duration-dependant
       HMM", 2016. Available at: https://github.com/davidspringer/Springer-Segmentation-Code   
    
    """

    # check inputs
    if envelope is None:
        raise TypeError("Please specify the signal's homomorphic envelope.")
    
    autocorrelation = np.correlate(envelope,envelope,mode='full')
    autocorrelation = autocorrelation[(autocorrelation.size)//2:]
    
    min_index = int(0.3*sampling_rate)
    max_index = int(1.5*sampling_rate)

    index = np.argmax(autocorrelation[min_index-1:max_index-1])
    true_index = index+min_index-1
    heartRate = 60/(true_index/sampling_rate)
    
    max_sys_duration = int(np.round(((60/heartRate)*sampling_rate)/2))
    min_sys_duration = int(np.round(0.2*sampling_rate))
    
    pos = np.argmax(autocorrelation[min_sys_duration-1:max_sys_duration-1])
    systolicTimeInterval = (min_sys_duration+pos)/sampling_rate
    

    return utils.ReturnTuple((heartRate,systolicTimeInterval),
                             ('heart_rate','systolic_time_interval'))

def identify_heart_sounds(beats = None, sampling_rate = 1000.):
    
    """Classify heart sound peaks as S1 or S2
     
    Parameters
    ----------
    beats : array
        Peaks of heart sounds
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
        
    Returns
    -------
    classification : array
        Classification of heart sound peaks. 1 is S1, 2 is S2
    
    """

    one_peak_ahead = np.roll(beats, -1)

    SS_intervals = (one_peak_ahead[0:-1] - beats[0:-1]) / sampling_rate
    
    # Initialize the vector to store the classification of the peaks:
    classification = np.zeros(len(beats))
        
    # Classify the peaks. 
    # Terrible algorithm, but good enough for now    
    for i in range(1,len(beats)-1):
        if SS_intervals[i-1] > SS_intervals[i]:
            classification[i] = 0
        else:
            classification[i] = 1
    classification[0] = int(not(classification[1]))
    classification[-1] = int(not(classification[-2]))    
    
    classification += 1    
        
    return utils.ReturnTuple((classification,), ('heart_sounds',))

def ecg_based_segmentation(pcg_signal=None, ecg_signal=None, sampling_rate=1000.0, show=False):
    """Assign state labels to PCG recording based on markers from simultaneous ECG signal.

    Adapted to Python from original MATLAB code written by David Springer, 2016 (C), for 
    comparison purposes in the paper [Springer15]_.
    Available at: https://github.com/davidspringer/Springer-Segmentation-Code 
    
    Heart sounds timing durations were obtained from [Schimdt10]_.
    
    Parameters
    ----------
    pcg_signal : array
        PCG signal to be segmented.
    ecg_signal : array
        Simultaneous ECG signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz) of both signals.
    show : bool, optional
        If True, show a plot with the segmented signal.
        
    Returns
    -------
    states : array
        State labels of PCG recording

    References
    ----------
    .. [Springer15] D.Springer, "Logistic Regression-HSMM-based Heart Sound Segmentation",
       IEEE Trans. Biomed. Eng., In Press, 2015.
    .. [Schimdt10] S. E. Schmidt et al., "Segmentation of heart sound recordings by a 
       duration-dependent hidden Markov model", Physiol. Meas., 2010    
    """
    
    # ensure numpy
    pcg_signal = np.array(pcg_signal)
    ecg_signal = np.array(ecg_signal)
    sampling_rate = float(sampling_rate)    

    # Compute homomorphic envelope to find peaks
    envelope, = homomorphic_filter(pcg_signal, sampling_rate=sampling_rate)
    envelope, = st.normalize(envelope)
        
    states = np.zeros((len(envelope),))
    
    # Extract locations for R peaks and end T-waves
    ecg_out = ecg.ecg(ecg_signal, sampling_rate=sampling_rate, show=False)
    rpeaks = ecg_out['rpeaks'].astype('int64')
    t_positions = ecg.getTPositions(ecg_out)
    t_ends = t_positions["T_end_positions"]
    
    #Timing durations from Schmidt:
    mean_S1 = 0.122*sampling_rate
    std_S1 = 0.022*sampling_rate
    mean_S2 = 0.092*sampling_rate
    std_S2 = 0.022*sampling_rate

    # Set duration from each R-peak to (R-peak+mean_S1) as first state S1.
    # Upper and lower bounds are set to avoid errors of searching outside size of the signal
    for peak in rpeaks:

        upper_bound = int(min([len(states), peak + mean_S1]))
        states[peak:min([upper_bound-1,len(states)])] = 1
    
    # Set S2 as state 3 depending on position of end T-wave in ECG
    for t_end in t_ends:
        
        # find search window of envelope: T-end +- mean+1std
        lower_bound = int(max([t_end - np.floor((mean_S2 + std_S2)),1]))
        upper_bound = int(min([len(states), np.ceil(t_end + np.floor(mean_S2 + std_S2))]))

        search_window = envelope[lower_bound-1:upper_bound]*(states[lower_bound-1:upper_bound]!=1).ravel()
        
        # Find the maximum value of the envelope in the search window:
        S2_index = np.argmax(search_window)
        
        # Find the actual index in the envelope of the maximum peak.
        # Make sure this has a max value of the length of the signal:
        S2_index = min([len(states),lower_bound+ S2_index-1])
        
        # Set the states to state 3, centered on the S2 peak, +- 1/2 of the
        # expected S2 sound duration.
        upper_bound = int(min([len(states), np.ceil(S2_index +((mean_S2)/2))]))
        states[int(max([np.ceil(S2_index - ((mean_S2)/2)),1]))-1:upper_bound] = 3
        
        # Set the spaces between state 3 and the next R peak as state 4:
        # Find the next rpeak after this S2. Exclude those that happened before by setting them to infinity.
        diffs = (rpeaks - t_end).astype(float)
        diffs[diffs<0] = np.inf
        
        # If the array is empty, then no S1s after this S2, so set to end of signal:
        if np.size(diffs[diffs<np.inf])==0:
            end_pos = len(states)
        else:
            # else, send the end position to the minimum diff
            index = np.argmin(diffs)
            end_pos = rpeaks[index]
            
        states[int(np.ceil(S2_index +((mean_S2 +(0*std_S2))/2))-1):end_pos] = 4
    
    # Set first and last sections of the signal (before first R-peak, and after last end T-wave)
    first_location_of_definite_state = np.argwhere(states!=0)[0][0]

    if first_location_of_definite_state > 0:
        if states[first_location_of_definite_state] == 1:
            states[0:first_location_of_definite_state] = 4
        
        if states[first_location_of_definite_state] == 3:
            states[0:first_location_of_definite_state+1] = 2    
    
    last_location_of_definite_state = np.argwhere(states!=0)[-1][0]
    
    if last_location_of_definite_state > 0:
        
        if states[last_location_of_definite_state] == 1:
            states[last_location_of_definite_state:] = 2
        
        if states[last_location_of_definite_state] == 1:
            states[last_location_of_definite_state:] = 4
    
    # Set everywhere else as state 2:        
    states[states == 0] = 2
    
    if show:
        fig, ax = plt.subplots(figsize=(15, 4))
        t = np.linspace(0,round(len(pcg_signal)/sampling_rate),len(pcg_signal))
    
        ax.plot(t, pcg_signal, color='black')
        
        # arrange samples into groups of equal elements (split arrays into individual states)
        time_intervals = np.split(t, np.where(np.diff(states) != 0)[0]+1)
        states_intervals = np.split(states, np.where(np.diff(states) != 0)[0]+1)
        
        colors = ["green","red","pink","blue"]
        y_lims = ax.get_ylim()

        for i in range(len(time_intervals)):
            ax.fill_between(time_intervals[i], y_lims[0], y_lims[1], color=colors[int(states_intervals[i][0]-1)], alpha=0.4)       
        
        legend_elements = [Patch(facecolor=colors[0],edgecolor=colors[0],alpha = 0.4,label='S1'),
                           Patch(facecolor=colors[1],edgecolor=colors[1],alpha = 0.4,label='Systole'),
                           Patch(facecolor=colors[2],edgecolor=colors[2],alpha = 0.4,label='S2'),
                           Patch(facecolor=colors[3],edgecolor=colors[3],alpha = 0.4,label='Diastole')]

        ax.legend(handles=legend_elements,bbox_to_anchor=(1.01, 1), loc='upper left')
        plt.tight_layout()
        plt.show()
        
    return utils.ReturnTuple((states,), 
                             ('states',))
                             



# -*- coding: utf-8 -*-
"""
biosppy.features.time
---------------------

This module provides methods to extract time features.

:copyright: (c) 2015-2023 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# Imports
# 3rd party
import numpy as np

# local
from .. import utils
from ..signals import tools as st
from .. import stats


def time(signal=None, sampling_rate=1000., include_diff=True):
    """Compute various time metrics describing the signal.

        Parameters
        ----------
        signal : array
            Input signal.
        sampling_rate : int, float, optional
            Sampling Rate (Hz).
        include_diff : bool, optional
            Whether to include the features of the signal's differences (first, second and absolute).

        Returns
        -------
        feats : ReturnTuple object
            Time features of the signal.

        Notes
        -----
        Besides the features directly extracted in this function, it also calls:
        - biosppy.signals.tools.signal_stats
        - biosppy.stats.quartiles
        - biosppy.stats.histogram
        - biosppy.features.time.hjorth_features

        """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    # initialize output
    feats = utils.ReturnTuple((), ())

    # basic stats
    signal_feats = st.signal_stats(signal)
    feats = feats.join(signal_feats)

    # quartile features
    quartile_feats = stats.quartiles(signal)
    feats = feats.join(quartile_feats)

    # number of maxima
    nb_maxima = st.find_extrema(signal, mode="max")
    feats = feats.append(len(nb_maxima['extrema']), 'nb_maxima')

    # number of minima
    nb_minima = st.find_extrema(signal, mode="min")
    feats = feats.append(len(nb_minima['extrema']), 'nb_minima')

    # autocorrelation sum
    autocorr_sum = np.sum(np.correlate(signal, signal, 'full'))
    feats = feats.append(autocorr_sum, 'autocorr_sum')

    # total energy
    total_energy = np.sum(np.abs(signal)**2)
    feats = feats.append(total_energy, 'total_energy')

    # histogram relative frequency
    hist_feats = stats.histogram(signal, normalize=True)
    feats = feats.join(hist_feats)

    # linear regression
    t_signal = np.arange(0, len(signal)) / sampling_rate
    linreg = stats.linear_regression(t_signal, signal, show=False)
    feats = feats.append(linreg['m'], 'linreg_slope')
    feats = feats.append(linreg['b'], 'linreg_intercept')

    # pearson correlation from linear regression
    linreg_pred = linreg['m'] * t_signal + linreg['b']
    pearson_feats = stats.pearson_correlation(signal, linreg_pred)
    feats = feats.append(pearson_feats['r'], 'pearson_r')

    # hjorth features
    hjorth_feats = hjorth_features(signal)
    feats = feats.join(hjorth_feats)

    # diff stats
    if include_diff:
        diff_feats = stats.diff_stats(signal, stats_only=True)
        feats = feats.join(diff_feats)

    return feats


def hjorth_features(signal=None):
    """Compute Hjorth mobility, complexity, chaos and hazard.

    Parameters
    ----------
    signal : array
        Input signal.

    Returns
    -------
    hjorth_mobility : float
        Hjorth mobility.
    hjorth_complexity : float
        Hjorth complexity.
    hjorth_chaos : float
        Hjorth chaos.
    hjorth_hazard : float
        Hjorth hazard.

    Notes
    -----
    Hjorth activity corresponds to the variance of the signal.

    """

    # helper functions
    def _hjorth_mobility(s):
        if np.var(s) == 0:
            return None
        return np.sqrt(np.var(np.diff(s)) / np.var(s))

    def _hjorth_complexity(s):
        mobility = _hjorth_mobility(s)
        if mobility is None or mobility == 0:
            return None
        return _hjorth_mobility(np.diff(s)) / mobility

    def _hjorth_chaos(s):
        complexity = _hjorth_complexity(s)
        if complexity is None or complexity == 0:
            return None
        return _hjorth_complexity(np.diff(s)) / complexity

    def _hjorth_hazard(s):
        chaos = _hjorth_chaos(s)
        if chaos is None or chaos == 0:
            return None
        return _hjorth_chaos(np.diff(s)) / chaos

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    # initialize output
    feats = utils.ReturnTuple((), ())

    # hjorth mobility
    signal_mobility = _hjorth_mobility(signal)
    feats = feats.append(signal_mobility, 'hjorth_mobility')
    if signal_mobility is None:
        print("Hjorth mobility is undefined. Returning None.")

    # hjorth complexity
    signal_complexity = _hjorth_complexity(signal)
    feats = feats.append(signal_complexity, 'hjorth_complexity')
    if signal_complexity is None:
        print("Hjorth complexity is undefined. Returning None.")

    # hjorth chaos
    signal_chaos = _hjorth_chaos(signal)
    feats = feats.append(signal_chaos, 'hjorth_chaos')
    if signal_chaos is None:
        print("Hjorth chaos is undefined. Returning None.")

    # hjorth hazard
    signal_hazard = _hjorth_hazard(signal)
    feats = feats.append(signal_hazard, 'hjorth_hazard')
    if signal_hazard is None:
        print("Hjorth hazard is undefined. Returning None.")

    return feats



# -*- coding: utf-8 -*-
"""
biosppy.plotting
----------------

This module provides utilities to plot data.

:copyright: (c) 2015-2023 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# Imports
# compat
from __future__ import absolute_import, division, print_function
from six.moves import range, zip
import six

# built-in
import os

# 3rd party
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.patches as patches
import matplotlib.lines as lines
import numpy as np

# local
from . import utils
from .signals import tools as st

# Globals
MAJOR_LW = 1.5
MED_LW = 1.25
MINOR_LW = 1.0
MAX_ROWS = 10
LOGOS_FOLDER = '.logos'
BIOSPPY_LOGO = 'biosppy.png'
SCIENTISST_LOGO = 'scientisst.png'

# matplotlib settings
plt.rcParams['text.color'] = '#495057'
plt.rcParams['axes.labelcolor'] = '#495057'
plt.rcParams['axes.labelsize'] = 10
plt.rcParams['axes.edgecolor'] = '#495057'
plt.rcParams['axes.facecolor'] = '#f8f9fa'
plt.rcParams['axes.titlesize'] = 11
plt.rcParams['xtick.color'] = '#495057'
plt.rcParams['ytick.color'] = '#495057'
plt.rcParams['grid.color'] = '#ebeef0'
plt.rcParams['axes.spines.top'] = False
plt.rcParams['axes.spines.right'] = False
plt.rcParams['axes.grid'] = True
plt.rcParams['grid.linestyle'] = '-'
plt.rcParams['grid.color'] = '#ebeef0'
plt.rcParams['legend.facecolor'] = 'white'
plt.rcParams['legend.framealpha'] = 0.75
plt.rcParams['legend.loc'] = 'upper right'
plt.rcParams['legend.fontsize'] = 8


def color_palette(idx):
    """Color palette to use throughout the biosppy package

    Parameters
    ----------
    idx: str or int
        identifier of color to use

    Returns
    -------
    color_id: str
        hexadecimal color code chosen
    """

    color_dict = {
        'blue': '#71a7cc',
        'dark-blue': '#184169',
        'light-blue': '#a5c8e0',

        'green': '#329352',
        'dark-green': '#154d28',
        'light-green': '#7ac77d',

        'red': '#D62839',
        'dark-red': '#B14343',
        'light-red': '#FF5A5F',

        'yellow': '#D19C2F',
        'dark-yellow': '#FFC300',
        'light-yellow': '#FFDD55',

        'violet': '#9D4EDD',
        'dark-violet': '#3C096C',
        'light-violet': '#E0AAFF',

        'orange': '#f3883e',
        'dark-orange': '#b2400a',
        'light-orange': '#f7d1ab',

        'grey': '#ADB5BD',
        'dark-grey': '#495057',
        'light-grey': '#E6E6E6'
    }

    if type(idx) == int:
        color_id = list(color_dict.values())[idx]
    else:
        if idx in color_dict.keys():
            color_id = color_dict[idx]
        else:
            raise ValueError(f'Please choose one color from {color_dict.keys()}'
                             f' or give an index')

    return color_id


def add_logo(fig):
    logos_folder = os.path.join(os.path.dirname(__file__), LOGOS_FOLDER)
    try:
        # add biosppy logo
        biosppy_logo = plt.imread(os.path.join(logos_folder, BIOSPPY_LOGO))
        logo = fig.add_axes([0.80, 0.02, 0.08, 0.08], anchor='SE')
        logo.imshow(biosppy_logo, alpha=0.5)
        logo.axis('off')

        # add scientisst logo
        scientisst_logo = plt.imread(os.path.join(logos_folder, SCIENTISST_LOGO))
        logo = fig.add_axes([0.90, 0.02, 0.08, 0.08], anchor='SE')
        logo.imshow(scientisst_logo, alpha=0.5)
        logo.axis('off')
    except:
        pass


def _plot_filter(b, a, sampling_rate=1000., nfreqs=4096, log_xscale=True,
                 ax=None):
    """Compute and plot the frequency response of a digital filter.

    Parameters
    ----------
    b : array
        Numerator coefficients.
    a : array
        Denominator coefficients.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    nfreqs : int, optional
        Number of frequency points to compute.
    log_xscale : bool, optional
        Whether to use log scale for x-axis.
    ax : axis, optional
        Plot Axis to use.

    Returns
    -------
    fig : Figure
        Figure object.

    """

    # compute frequency response
    freqs, resp = st._filter_resp(b, a,
                                  sampling_rate=sampling_rate,
                                  nfreqs=nfreqs)

    # plot
    if ax is None:
        fig = plt.figure(figsize=(8, 4))
        ax = fig.add_subplot(211)
        ax2 = fig.add_subplot(212)
    else:
        ax2 = ax.twinx()

    # set layout
    fig.subplots_adjust(top=0.88, bottom=0.17, hspace=0.2, left=0.16,
                        right=0.94)

    # title
    fig.suptitle('Filter Frequency Response', fontsize=12, fontweight='bold')

    # amplitude
    pwr = 20. * np.log10(np.abs(resp))
    if log_xscale:
        ax.semilogx(freqs, pwr, color=color_palette('blue'),
                    linewidth=MAJOR_LW)
    else:
        ax.plot(freqs, pwr, color=color_palette('blue'), linewidth=MAJOR_LW)
    ax.set_ylabel('Amplitude (dB)')
    ax.grid(True)
    ax.tick_params(axis='x', which='both', bottom=False, top=False,
                   labelbottom=False)
    ax.spines['bottom'].set_visible(False)

    # phase
    angles = np.unwrap(np.angle(resp)) * 180. / np.pi  # radians to degrees
    if log_xscale:
        ax2.semilogx(freqs, angles, color=color_palette('dark-blue'),
                     linewidth=MAJOR_LW)
    else:
        ax2.plot(freqs, angles, color=color_palette('dark-blue'),
                 linewidth=MAJOR_LW)
    ax2.set_ylabel('Angle (degrees)')
    ax2.set_xlabel('Frequency (Hz)')
    ax2.grid(True)

    # align y-axis labels
    fig.align_ylabels()

    # add logo
    add_logo(fig)

    return fig


def plot_filter(ftype='FIR',
                band='lowpass',
                order=None,
                frequency=None,
                sampling_rate=1000.,
                log_xscale=True,
                path=None,
                show=True, **kwargs):
    """Plot the frequency response of the filter specified with the given
    parameters.

    Parameters
    ----------
    ftype : str
        Filter type:
            * Finite Impulse Response filter ('FIR');
            * Butterworth filter ('butter');
            * Chebyshev filters ('cheby1', 'cheby2');
            * Elliptic filter ('ellip');
            * Bessel filter ('bessel').
    band : str
        Band type:
            * Low-pass filter ('lowpass');
            * High-pass filter ('highpass');
            * Band-pass filter ('bandpass');
            * Band-stop filter ('bandstop').
    order : int
        Order of the filter.
    frequency : int, float, list, array
        Cutoff frequencies; format depends on type of band:
            * 'lowpass' or 'bandpass': single frequency;
            * 'bandpass' or 'bandstop': pair of frequencies.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    log_xscale : bool, optional
        Whether to use log scale for x-axis.
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show the plot immediately.
    ``**kwargs`` : dict, optional
        Additional keyword arguments are passed to the underlying
        scipy.signal function.

    """

    # get filter
    b, a = st.get_filter(ftype=ftype,
                         band=band,
                         order=order,
                         frequency=frequency,
                         sampling_rate=sampling_rate, **kwargs)

    # plot
    fig = _plot_filter(b, a, sampling_rate, log_xscale=log_xscale)

    # save to file
    if path is not None:
        path = utils.normpath(path)
        root, ext = os.path.splitext(path)
        ext = ext.lower()
        if ext not in ['png', 'jpg']:
            path = root + '.png'

        fig.savefig(path, dpi=200, bbox_inches='tight')

    # show
    if show:
        plt.show()
    else:
        # close
        plt.close(fig)


def plot_spectrum(signal=None, sampling_rate=1000., path=None, show=True):
    """Plot the power spectrum of a signal (one-sided).

    Parameters
    ----------
    signal : array
        Input signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show the plot immediately.

    """

    freqs, power = st.power_spectrum(signal, sampling_rate,
                                     pad=0,
                                     pow2=False,
                                     decibel=True)

    fig = plt.figure()
    ax = fig.add_subplot(111)

    ax.plot(freqs, power, linewidth=MAJOR_LW)
    ax.set_xlabel('Frequency (Hz)')
    ax.set_ylabel('Power (dB)')
    ax.grid()

    # make layout tight
    fig.tight_layout()

    # save to file
    if path is not None:
        path = utils.normpath(path)
        root, ext = os.path.splitext(path)
        ext = ext.lower()
        if ext not in ['png', 'jpg']:
            path = root + '.png'

        fig.savefig(path, dpi=200, bbox_inches='tight')

    # show
    if show:
        plt.show()
    else:
        # close
        plt.close(fig)


def plot_acc(ts=None,
             raw=None,
             vm=None,
             sm=None,
             units=None,
             path=None,
             show=False):
    """Create a summary plot from the output of signals.acc.acc.

    Parameters
    ----------
    ts : array
        Signal time axis reference (seconds).
    raw : array
        Raw ACC signal.
    vm : array
        Vector Magnitude feature of the signal.
    sm : array
        Signal Magnitude feature of the signal
    units : str, optional
        Units of the vertical axis. If provided, the plot title will include
        the units information. Default is None.
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show the plot immediately.

    """

    raw_t = np.transpose(raw)
    acc_x, acc_y, acc_z = raw_t[0], raw_t[1], raw_t[2]

    fig = plt.figure(figsize=(10, 5))
    fig.suptitle('ACC Summary', fontsize=12, fontweight='bold')
    gs = gridspec.GridSpec(6, 2)
    fig.subplots_adjust(top=0.85, hspace=0.7, wspace=0.34, left=0.13,
                        right=0.96, bottom=0.18)

    # raw signal (acc_x)
    ax1 = fig.add_subplot(gs[:2, 0])
    ax1.set_title("Signal")

    ax1.plot(ts, acc_x, linewidth=MINOR_LW, label='X',
             color=color_palette('dark-blue'))

    ax1.legend(loc='upper right')
    ax1.tick_params(axis='x', which='both', bottom=False, top=False,
                    labelbottom=False)
    ax1.spines['bottom'].set_visible(False)

    # raw signal (acc_y)
    ax2 = fig.add_subplot(gs[2:4, 0], sharex=ax1)

    ax2.plot(ts, acc_y, linewidth=MINOR_LW, label='Y',
             color=color_palette('blue'))

    ax2.set_ylabel('Amplitude' if units is None else 'Amplitude ($%s$)' % units)
    ax2.legend(loc='upper right')
    ax2.tick_params(axis='x', which='both', bottom=False, top=False,
                    labelbottom=False)
    ax2.spines['bottom'].set_visible(False)

    # raw signal (acc_z)
    ax3 = fig.add_subplot(gs[4:, 0], sharex=ax1)

    ax3.plot(ts, acc_z, linewidth=MINOR_LW, label='Z',
             color=color_palette('light-blue'))

    ax3.set_xlabel('Time (s)')
    ax3.legend(loc='upper right')

    # vector magnitude
    ax4 = fig.add_subplot(gs[:3, 1], sharex=ax1)
    ax4.set_title('Features')

    ax4.plot(ts, vm, linewidth=MINOR_LW, label='Vector Magnitude',
             color=color_palette('orange'))

    ax4.set_ylabel('Amplitude' if units is None else 'Amplitude ($%s$)' % units)
    ax4.legend(loc='upper right')
    ax4.tick_params(axis='x', which='both', bottom=False, top=False,
                    labelbottom=False)
    ax4.spines['bottom'].set_visible(False)

    # signal magnitude
    ax5 = fig.add_subplot(gs[3:, 1], sharex=ax1)

    ax5.plot(ts, sm, linewidth=MINOR_LW, label='Signal Magnitude',
             color=color_palette('light-orange'))

    ax5.set_ylabel('Amplitude' if units is None else 'Amplitude (%s)' % units)
    ax5.set_xlabel('Time (s)')
    ax5.legend(loc='upper right')

    # align y-axis labels
    fig.align_ylabels([ax1, ax2, ax3])
    fig.align_ylabels([ax4, ax5])

    # add logo
    add_logo(fig)

    # save to file
    if path is not None:
        path = utils.normpath(path)
        root, ext = os.path.splitext(path)
        ext = ext.lower()
        if ext not in ['png', 'jpg']:
            path = root + '.png'

        fig.savefig(path, dpi=200, bbox_inches='tight')

    # show
    if show:
        plt.show()
    else:
        # close
        plt.close(fig)


def plot_ppg(ts=None,
             raw=None,
             filtered=None,
             peaks=None,
             templates_ts=None,
             templates=None,
             heart_rate_ts=None,
             heart_rate=None,
             units=None,
             path=None,
             show=False):
    """Create a summary plot from the output of signals.ppg.ppg.

    Parameters
    ----------
    ts : array
        Signal time axis reference (seconds).
    raw : array
        Raw PPG signal.
    filtered : array
        Filtered PPG signal.
    peaks : array
        Indices of PPG pulse peaks.
    templates_ts : array
        Templates time axis reference (seconds).
    templates : array
        Extracted PPG templates.
    heart_rate_ts : array
        Heart rate time axis reference (seconds).
    heart_rate : array
        Instantaneous heart rate (bpm).
    units : str, optional
        Units of the vertical axis. If provided, the plot title will include
        the units information. Default is None.
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show the plot immediately.

    """

    fig = plt.figure(figsize=(10, 5))
    fig.suptitle('PPG Summary', fontsize=12, fontweight='bold')
    gs = gridspec.GridSpec(6, 2)
    fig.subplots_adjust(top=0.88, bottom=0.12, hspace=0.9, wspace=0.3,
                        left=0.1, right=0.96)

    # raw signal
    ax1 = fig.add_subplot(gs[:2, 0])
    ax1.set_title('Signal')

    ax1.plot(ts, raw, linewidth=MED_LW, label='Raw',
             color=color_palette('light-blue'), alpha=0.6)
    ax1.plot(ts, filtered+np.mean(raw), linewidth=MINOR_LW, label='Filtered',
             color=color_palette('blue'))

    ax1.set_ylabel('Amplitude' if units is None else 'Amplitude (%s)' % units)
    ax1.legend()
    ax1.tick_params(axis='x', which='both', bottom=False, top=False,
                    labelbottom=False)
    ax1.spines['bottom'].set_visible(False)

    #  signal with onsets
    ax2 = fig.add_subplot(gs[2:4, 0], sharex=ax1)
    ax2.set_title('Systolic Peak Detection')

    ymin = np.min(filtered)
    ymax = np.max(filtered)
    alpha = 0.1 * (ymax - ymin)
    ymax += alpha
    ymin -= alpha

    ax2.plot(ts, filtered, linewidth=MINOR_LW, color=color_palette('blue'))
    ax2.plot(ts[peaks], filtered[peaks], ls='None', marker='x',
             color=color_palette('dark-red'), label='Peaks')

    ax2.set_ylabel('Amplitude' if units is None else 'Amplitude (%s)' % units)
    ax2.legend()

    ax2.tick_params(axis='x', which='both', bottom=False, top=False,
                    labelbottom=False)
    ax2.spines['bottom'].set_visible(False)

    # heart rate
    ax3 = fig.add_subplot(gs[4:, 0], sharex=ax1)
    ax3.set_title('Heart Rate')

    ax3.plot(heart_rate_ts, heart_rate, linewidth=MAJOR_LW, label='Heart Rate',
             color=color_palette('blue'))

    ax3.set_xlabel('Time (s)')
    ax3.set_ylabel('Heart Rate (bpm)')

    # align y-axis labels
    fig.align_ylabels([ax1, ax2, ax3])

    # templates
    ax4 = fig.add_subplot(gs[1:5, 1])

    ax4.plot(templates_ts, templates, linewidth=MINOR_LW, alpha=0.5,
             color=color_palette('blue'))

    ax4.set_xlabel('Time (s)')
    ax4.set_ylabel('Amplitude' if units is None else 'Amplitude (%s)' % units)
    ax4.set_title('Templates')

    add_logo(fig)

    # save to file
    if path is not None:
        path = utils.normpath(path)
        root, ext = os.path.splitext(path)
        ext = ext.lower()
        if ext not in ['png', 'jpg']:
            path = root + '.png'

        fig.savefig(path, dpi=200, bbox_inches='tight')

    # show
    if show:
        plt.show()
    else:
        # close
        plt.close(fig)


def plot_bvp(ts=None,
             raw=None,
             filtered=None,
             onsets=None,
             heart_rate_ts=None,
             heart_rate=None,
             path=None,
             show=False):
    """Create a summary plot from the output of signals.bvp.bvp.

    Parameters
    ----------
    ts : array
        Signal time axis reference (seconds).
    raw : array
        Raw BVP signal.
    filtered : array
        Filtered BVP signal.
    onsets : array
        Indices of BVP pulse onsets.
    heart_rate_ts : array
        Heart rate time axis reference (seconds).
    heart_rate : array
        Instantaneous heart rate (bpm).
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show the plot immediately.

    """

    fig = plt.figure()
    fig.suptitle('BVP Summary')

    # raw signal
    ax1 = fig.add_subplot(311)

    ax1.plot(ts, raw, linewidth=MAJOR_LW, label='Raw')

    ax1.set_ylabel('Amplitude')
    ax1.legend()
    ax1.grid()

    # filtered signal with onsets
    ax2 = fig.add_subplot(312, sharex=ax1)

    ymin = np.min(filtered)
    ymax = np.max(filtered)
    alpha = 0.1 * (ymax - ymin)
    ymax += alpha
    ymin -= alpha

    ax2.plot(ts, filtered, linewidth=MAJOR_LW, label='Filtered')
    ax2.vlines(ts[onsets], ymin, ymax,
               color='m',
               linewidth=MINOR_LW,
               label='Onsets')

    ax2.set_ylabel('Amplitude')
    ax2.legend()
    ax2.grid()

    # heart rate
    ax3 = fig.add_subplot(313, sharex=ax1)

    ax3.plot(heart_rate_ts, heart_rate, linewidth=MAJOR_LW, label='Heart Rate')

    ax3.set_xlabel('Time (s)')
    ax3.set_ylabel('Heart Rate (bpm)')
    ax3.legend()
    ax3.grid()

    # make layout tight
    fig.tight_layout()

    # save to file
    if path is not None:
        path = utils.normpath(path)
        root, ext = os.path.splitext(path)
        ext = ext.lower()
        if ext not in ['png', 'jpg']:
            path = root + '.png'

        fig.savefig(path, dpi=200, bbox_inches='tight')

    # show
    if show:
        plt.show()
    else:
        # close
        plt.close(fig)


def plot_abp(ts=None,
             raw=None,
             filtered=None,
             onsets=None,
             heart_rate_ts=None,
             heart_rate=None,
             path=None,
             show=False):
    """Create a summary plot from the output of signals.abp.abp.

    Parameters
    ----------
    ts : array
        Signal time axis reference (seconds).
    raw : array
        Raw ABP signal.
    filtered : array
        Filtered ABP signal.
    onsets : array
        Indices of ABP pulse onsets.
    heart_rate_ts : array
        Heart rate time axis reference (seconds).
    heart_rate : array
        Instantaneous heart rate (bpm).
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show the plot immediately.

    """

    fig = plt.figure()
    fig.suptitle('ABP Summary')

    # raw signal
    ax1 = fig.add_subplot(311)

    ax1.plot(ts, raw, linewidth=MAJOR_LW, label='Raw')

    ax1.set_ylabel('Amplitude')
    ax1.legend()
    ax1.grid()

    # filtered signal with onsets
    ax2 = fig.add_subplot(312, sharex=ax1)

    ymin = np.min(filtered)
    ymax = np.max(filtered)
    alpha = 0.1 * (ymax - ymin)
    ymax += alpha
    ymin -= alpha

    ax2.plot(ts, filtered, linewidth=MAJOR_LW, label='Filtered')
    ax2.vlines(ts[onsets], ymin, ymax,
               color='m',
               linewidth=MINOR_LW,
               label='Onsets')

    ax2.set_ylabel('Amplitude')
    ax2.legend()
    ax2.grid()

    # heart rate
    ax3 = fig.add_subplot(313, sharex=ax1)

    ax3.plot(heart_rate_ts, heart_rate, linewidth=MAJOR_LW, label='Heart Rate')

    ax3.set_xlabel('Time (s)')
    ax3.set_ylabel('Heart Rate (bpm)')
    ax3.legend()
    ax3.grid()

    # make layout tight
    fig.tight_layout()

    # save to file
    if path is not None:
        path = utils.normpath(path)
        root, ext = os.path.splitext(path)
        ext = ext.lower()
        if ext not in ['png', 'jpg']:
            path = root + '.png'

        fig.savefig(path, dpi=200, bbox_inches='tight')

    # show
    if show:
        plt.show()
    else:
        # close
        plt.close(fig)


def plot_eda(ts=None,
             raw=None,
             filtered=None,
             edr=None,
             edl=None,
             onsets=None,
             peaks=None,
             amplitudes=None,
             units=None,
             path=None,
             show=False):
    """Create a summary plot from the output of signals.eda.eda.

    Parameters
    ----------
    ts : array
        Signal time axis reference (seconds).
    raw : array
        Raw EDA signal.
    filtered : array
        Filtered EDA signal.
    edr : array
        Electrodermal response signal.
    edl : array
        Electrodermal level signal.
    onsets : array
        Events onsets indices.
    peaks : array
        Events peaks indices.
    amplitudes : array
        Amplitudes location indices.
    units : str, optional
        Units of the vertical axis. If provided, the plot title will include
        the units information. Default is None.
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show the plot immediately.

    """

    fig = plt.figure(figsize=(10, 5))
    fig.suptitle('EDA Summary', fontsize=14, fontweight='bold')
    gs = gridspec.GridSpec(6, 2)
    fig.subplots_adjust(top=0.85, hspace=0.7, wspace=0.34, left=0.13,
                        right=0.96, bottom=0.18)

    # signal
    ax1 = fig.add_subplot(gs[:2, 0])

    ax1.plot(ts, raw, linewidth=MED_LW, label='Raw',
             color=color_palette('light-blue'), alpha=0.6)
    ax1.plot(ts, filtered, linewidth=MINOR_LW, label='Filtered',
             color=color_palette('blue'))

    ax1.set_ylabel('Amplitude' if units is None else 'Amplitude \n (%s)' % units)
    ax1.set_title('Signal')
    ax1.legend()

    ax1.tick_params(axis='x', which='both', bottom=False, top=False,
                    labelbottom=False)
    ax1.spines['bottom'].set_visible(False)

    # event detection
    ax2 = fig.add_subplot(gs[2:4, 0], sharex=ax1, sharey=ax1)

    ymin = np.min(filtered)
    ymax = np.max(filtered)
    alpha = 0.1 * (ymax - ymin)
    ymax += alpha
    ymin -= alpha

    ax2.plot(ts, filtered, linewidth=MINOR_LW, color=color_palette('blue'))
    ax2.scatter(ts[onsets], filtered[onsets], marker='|', lw=1, s=100,
                color=color_palette('dark-red'), label='Onsets', zorder=3)
    ax2.scatter(ts[peaks], filtered[peaks], marker='x',
                color=color_palette('dark-red'), label='Peaks', zorder=3)

    ax2.set_ylabel('Amplitude' if units is None else 'Amplitude \n (%s)' % units)
    ax2.set_title('Event Detection')
    ax2.legend()

    ax2.tick_params(axis='x', which='both', bottom=False, top=False,
                    labelbottom=False)
    ax2.spines['bottom'].set_visible(False)

    # amplitudes
    ax3 = fig.add_subplot(gs[4:, 0], sharex=ax1)

    ax3.plot(ts[peaks], amplitudes, linewidth=MAJOR_LW, color=color_palette('blue'), label='Amplitude')

    ax3.set_xlabel('Time (s)')
    ax3.set_ylabel('Amplitude' if units is None else 'Amplitude \n (%s)' % units)
    ax3.set_title('Event Amplitudes')

    # align y axis labels
    fig.align_ylabels()

    # decomposition EDL
    ax4 = fig.add_subplot(gs[0:3, 1], sharex=ax1, sharey=ax1)

    ax4.plot(ts, filtered, linewidth=MAJOR_LW, color=color_palette('light-blue'))
    ax4.plot(ts, edl, linewidth=MINOR_LW, color=color_palette('dark-orange'),
             label="EDL")

    ax4.set_ylabel('Amplitude' if units is None else 'Amplitude \n (%s)' % units)
    ax4.set_title('EDA Decomposition')
    ax4.legend()

    ax4.tick_params(axis='x', which='both', bottom=False, top=False,
                    labelbottom=False)
    ax4.spines['bottom'].set_visible(False)

    # decomposition EDR
    ax5 = fig.add_subplot(gs[3:, 1], sharex=ax1)

    ax5.plot(ts[1:], edr, linewidth=MINOR_LW,
             color=color_palette('dark-orange'), label="EDR")

    ax5.set_ylabel('Amplitude' if units is None else 'Amplitude \n (%s)' % units)
    ax5.set_xlabel('Time (s)')
    ax5.legend()

    # logo
    add_logo(fig)

    # save to file
    if path is not None:
        path = utils.normpath(path)
        root, ext = os.path.splitext(path)
        ext = ext.lower()
        if ext not in ['png', 'jpg']:
            path = root + '.png'

        fig.savefig(path, dpi=200, bbox_inches='tight')

    # show
    if show:
        plt.show()
    else:
        # close
        plt.close(fig)


def plot_emg(ts=None,
             sampling_rate=None,
             raw=None,
             filtered=None,
             onsets=None,
             processed=None,
             units=None,
             path=None,
             show=False):
    """Create a summary plot from the output of signals.emg.emg.

    Parameters
    ----------
    ts : array
        Signal time axis reference (seconds).
    sampling_rate : int, float
        Sampling frequency (Hz).
    raw : array
        Raw EMG signal.
    filtered : array
        Filtered EMG signal.
    onsets : array
        Indices of EMG pulse onsets.
    processed : array, optional
        Processed EMG signal according to the chosen onset detector.
    units : str, optional
        Units of the vertical axis. If provided, the plot title will include
        the units information. Default is None.
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show the plot immediately.

    """

    fig = plt.figure(figsize=(10, 5))
    fig.suptitle('EMG Summary', fontsize=12, fontweight='bold')
    fig.subplots_adjust(top=0.85, hspace=0.25, wspace=0.34, left=0.1,
                        right=0.96, bottom=0.18)

    if processed is not None:
        ax1 = fig.add_subplot(311)
        ax2 = fig.add_subplot(312, sharex=ax1)
        ax3 = fig.add_subplot(313)

        # processed signal
        L = len(processed)
        T = (L - 1) / sampling_rate
        ts_processed = np.linspace(0, T, L, endpoint=True)
        ax3.plot(ts_processed, processed,
                 linewidth=MINOR_LW,
                 label='Processed',
                 color=color_palette('blue'))
        ax3.set_xlabel('Time (s)')
        ax3.set_ylabel('Amplitude')
        ax3.legend(loc='upper right')
    else:
        ax1 = fig.add_subplot(211)
        ax2 = fig.add_subplot(212, sharex=ax1)

    # signal
    ax1.set_title('Signal')
    ax1.plot(ts, raw, linewidth=MED_LW, label='Raw', alpha=0.6,
             color=color_palette('light-blue'))
    ax1.plot(ts, filtered+np.mean(raw), linewidth=MINOR_LW, label='Filtered',
             color=color_palette('blue'))

    ax1.set_ylabel('Amplitude' if units is None else 'Amplitude (%s)' % units)
    ax1.legend(loc='upper right')
    ax1.tick_params(axis='x', which='both', bottom=False, top=False,
                    labelbottom=False)
    ax1.spines['bottom'].set_visible(False)

    # filtered signal with onsets
    ymin = np.min(filtered)
    ymax = np.max(filtered)
    alpha = 0.1 * (ymax - ymin)
    ymax += alpha
    ymin -= alpha

    ax2.set_title('Event Detection')
    ax2.plot(ts, filtered, linewidth=MINOR_LW, color=color_palette('blue'))
    ax2.vlines(ts[onsets], ymin, ymax,
               color=color_palette('orange'),
               linewidth=MINOR_LW,
               label='Onsets')

    ax2.set_xlabel('Time (s)')
    ax2.set_ylabel('Amplitude' if units is None else 'Amplitude (%s)' % units)
    ax2.legend(loc='upper right')

    # align y axis labels
    fig.align_ylabels()

    # add logo
    add_logo(fig)

    # save to file
    if path is not None:
        path = utils.normpath(path)
        root, ext = os.path.splitext(path)
        ext = ext.lower()
        if ext not in ['png', 'jpg']:
            path = root + '.png'

        fig.savefig(path, dpi=200, bbox_inches='tight')

    # show
    if show:
        plt.show()
    else:
        # close
        plt.close(fig)


def plot_resp(ts=None,
              raw=None,
              filtered=None,
              zeros=None,
              resp_rate_ts=None,
              resp_rate=None,
              units=None,
              path=None,
              show=False):
    """Create a summary plot from the output of signals.ppg.ppg.

    Parameters
    ----------
    ts : array
        Signal time axis reference (seconds).
    raw : array
        Raw Resp signal.
    filtered : array
        Filtered Resp signal.
    zeros : array
        Indices of Respiration zero crossings.
    resp_rate_ts : array
        Respiration rate time axis reference (seconds).
    resp_rate : array
        Instantaneous respiration rate (Hz).
    units : str, optional
        Units of the vertical axis. If provided, the plot title will include
        the units information. Default is None.
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show the plot immediately.

    """

    fig = plt.figure(figsize=(10, 5))
    fig.suptitle('RESP Summary', fontsize=12, fontweight='bold')
    fig.subplots_adjust(top=0.85, hspace=0.35, wspace=0.34, left=0.13,
                        right=0.96, bottom=0.18)

    # raw signal
    ax1 = fig.add_subplot(311)
    ax1.set_title('Signal')

    ax1.plot(ts, raw, linewidth=MED_LW, label='Raw', alpha=0.6,
             color=color_palette('light-blue'))
    ax1.plot(ts, filtered+np.mean(raw), linewidth=MINOR_LW, label='Filtered',
                color=color_palette('blue'))

    ax1.set_ylabel('Amplitude' if units is None else 'Amplitude \n (%s)' % units)
    ax1.legend(loc='upper right')
    ax1.tick_params(axis='x', which='both', bottom=False, top=False,
                    labelbottom=False)
    ax1.spines['bottom'].set_visible(False)

    # filtered signal with zeros
    ax2 = fig.add_subplot(312, sharex=ax1)

    ymin = np.min(filtered)
    ymax = np.max(filtered)
    alpha = 0.1 * (ymax - ymin)
    ymax += alpha
    ymin -= alpha

    ax2.plot(ts, filtered, linewidth=MINOR_LW, color=color_palette('blue'))
    ax2.vlines(ts[zeros], ymin, ymax,
               color=color_palette('dark-red'),
               linewidth=MINOR_LW,
               label='Zero crossings')

    ax2.set_ylabel('Amplitude' if units is None else 'Amplitude \n (%s)' % units)
    ax2.legend(loc='upper right')
    ax2.tick_params(axis='x', which='both', bottom=False, top=False,
                    labelbottom=False)
    ax2.spines['bottom'].set_visible(False)

    # respiration rate
    ax3 = fig.add_subplot(313, sharex=ax1)
    ax3.set_title('Respiration Rate')

    ax3.plot(resp_rate_ts, resp_rate, linewidth=MAJOR_LW,
             color=color_palette('blue'))

    ax3.set_xlabel('Time (s)')
    ax3.set_ylabel('Respiration Rate \n(Hz)')

    # align y axes labels
    fig.align_ylabels()

    # add logo
    add_logo(fig)

    # save to file
    if path is not None:
        path = utils.normpath(path)
        root, ext = os.path.splitext(path)
        ext = ext.lower()
        if ext not in ['png', 'jpg']:
            path = root + '.png'

        fig.savefig(path, dpi=200, bbox_inches='tight')

    # show
    if show:
        plt.show()
    else:
        # close
        plt.close(fig)


def plot_eeg(ts=None,
             raw=None,
             filtered=None,
             labels=None,
             features_ts=None,
             theta=None,
             alpha_low=None,
             alpha_high=None,
             beta=None,
             gamma=None,
             plf_pairs=None,
             plf=None,
             path=None,
             show=False):
    """Create a summary plot from the output of signals.eeg.eeg.

    Parameters
    ----------
    ts : array
        Signal time axis reference (seconds).
    raw : array
        Raw EEG signal.
    filtered : array
        Filtered EEG signal.
    labels : list
        Channel labels.
    features_ts : array
        Features time axis reference (seconds).
    theta : array
        Average power in the 4 to 8 Hz frequency band; each column is one
        EEG channel.
    alpha_low : array
        Average power in the 8 to 10 Hz frequency band; each column is one
        EEG channel.
    alpha_high : array
        Average power in the 10 to 13 Hz frequency band; each column is one
        EEG channel.
    beta : array
        Average power in the 13 to 25 Hz frequency band; each column is one
        EEG channel.
    gamma : array
        Average power in the 25 to 40 Hz frequency band; each column is one
        EEG channel.
    plf_pairs : list
        PLF pair indices.
    plf : array
        PLF matrix; each column is a channel pair.
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show the plot immediately.

    """

    nrows = MAX_ROWS
    alpha = 2.

    # Get number of channels
    nch = raw.shape[1]

    figs = []

    # raw
    fig = _plot_multichannel(ts=ts,
                             signal=raw,
                             labels=labels,
                             nrows=nrows,
                             alpha=alpha,
                             title='EEG Summary - Raw',
                             xlabel='Time (s)',
                             ylabel='Amplitude')
    figs.append(('_Raw', fig))

    # filtered
    fig = _plot_multichannel(ts=ts,
                             signal=filtered,
                             labels=labels,
                             nrows=nrows,
                             alpha=alpha,
                             title='EEG Summary - Filtered',
                             xlabel='Time (s)',
                             ylabel='Amplitude')
    figs.append(('_Filtered', fig))

    # band-power
    names = ('Theta Band', 'Lower Alpha Band', 'Higher Alpha Band',
             'Beta Band', 'Gamma Band')
    args = (theta, alpha_low, alpha_high, beta, gamma)
    for n, a in zip(names, args):
        fig = _plot_multichannel(ts=features_ts,
                                 signal=a,
                                 labels=labels,
                                 nrows=nrows,
                                 alpha=alpha,
                                 title='EEG Summary - %s' % n,
                                 xlabel='Time (s)',
                                 ylabel='Power')
        figs.append(('_' + n.replace(' ', '_'), fig))

    # Only plot/compute plf if there is more than one channel
    if nch > 1:
        # PLF
        plf_labels = ['%s vs %s' % (labels[p[0]], labels[p[1]]) for p in plf_pairs]
        fig = _plot_multichannel(ts=features_ts,
                                 signal=plf,
                                 labels=plf_labels,
                                 nrows=nrows,
                                 alpha=alpha,
                                 title='EEG Summary - Phase-Locking Factor',
                                 xlabel='Time (s)',
                                 ylabel='PLF')
        figs.append(('_PLF', fig))

    # save to file
    if path is not None:
        path = utils.normpath(path)
        root, ext = os.path.splitext(path)
        ext = ext.lower()
        if ext not in ['png', 'jpg']:
            ext = '.png'

        for n, fig in figs:
            path = root + n + ext
            fig.savefig(path, dpi=200, bbox_inches='tight')

    # show
    if show:
        plt.show()
    else:
        # close
        for _, fig in figs:
            plt.close(fig)


def _yscaling(signal=None, alpha=1.5):
    """Get y axis limits for a signal with scaling.

    Parameters
    ----------
    signal : array
        Input signal.
    alpha : float, optional
        Scaling factor.

    Returns
    -------
    ymin : float
        Minimum y value.
    ymax : float
        Maximum y value.

    """

    mi = np.min(signal)
    m = np.mean(signal)
    mx = np.max(signal)

    if mi == mx:
        ymin = m - 1
        ymax = m + 1
    else:
        ymin = m - alpha * (m - mi)
        ymax = m + alpha * (mx - m)

    return ymin, ymax


def _plot_multichannel(ts=None,
                       signal=None,
                       labels=None,
                       nrows=10,
                       alpha=2.,
                       title=None,
                       xlabel=None,
                       ylabel=None):
    """Plot a multi-channel signal.

    Parameters
    ----------
    ts : array
        Signal time axis reference (seconds).
    signal : array
        Multi-channel signal; each column is one channel.
    labels : list, optional
        Channel labels.
    nrows : int, optional
        Maximum number of rows to use.
    alpha : float, optional
        Scaling factor for y axis.
    title : str, optional
        Plot title.
    xlabel : str, optional
        Label for x axis.
    ylabel : str, optional
        Label for y axis.

    Returns
    -------
    fig : Figure
        Figure object.

    """

    # ensure numpy
    signal = np.array(signal)
    nch = signal.shape[1]

    # check labels
    if labels is None:
        labels = ['Ch. %d' % i for i in range(nch)]

    if nch < nrows:
        nrows = nch

    ncols = int(np.ceil(nch / float(nrows)))

    fig = plt.figure()

    # title
    if title is not None:
        fig.suptitle(title)

    gs = gridspec.GridSpec(nrows, ncols, hspace=0, wspace=0.2)

    # reference axes
    ax0 = fig.add_subplot(gs[0, 0])
    ax0.plot(ts, signal[:, 0], linewidth=MAJOR_LW, label=labels[0])
    ymin, ymax = _yscaling(signal[:, 0], alpha=alpha)
    ax0.set_ylim(ymin, ymax)
    ax0.legend()
    ax0.grid()
    axs = {(0, 0): ax0}

    for i in range(1, nch - 1):
        a = i % nrows
        b = int(np.floor(i / float(nrows)))
        ax = fig.add_subplot(gs[a, b], sharex=ax0)
        axs[(a, b)] = ax

        ax.plot(ts, signal[:, i], linewidth=MAJOR_LW, label=labels[i])
        ymin, ymax = _yscaling(signal[:, i], alpha=alpha)
        ax.set_ylim(ymin, ymax)
        ax.legend()
        ax.grid()

    # last plot
    i = nch - 1
    a = i % nrows
    b = int(np.floor(i / float(nrows)))
    ax = fig.add_subplot(gs[a, b], sharex=ax0)
    axs[(a, b)] = ax

    ax.plot(ts, signal[:, -1], linewidth=MAJOR_LW, label=labels[-1])
    ymin, ymax = _yscaling(signal[:, -1], alpha=alpha)
    ax.set_ylim(ymin, ymax)
    ax.legend()
    ax.grid()

    if xlabel is not None:
        ax.set_xlabel(xlabel)

        for b in range(0, ncols - 1):
            a = nrows - 1
            ax = axs[(a, b)]
            ax.set_xlabel(xlabel)

    if ylabel is not None:
        # middle left
        a = nrows // 2
        ax = axs[(a, 0)]
        ax.set_ylabel(ylabel)

    # make layout tight
    gs.tight_layout(fig)

    return fig


def plot_ecg(ts=None,
             raw=None,
             filtered=None,
             rpeaks=None,
             templates_ts=None,
             templates=None,
             heart_rate_ts=None,
             heart_rate=None,
             units=None,
             path=None,
             show=False):
    """Create a summary plot from the output of signals.ecg.ecg.

    Parameters
    ----------
    ts : array
        Signal time axis reference (seconds).
    raw : array
        Raw ECG signal.
    filtered : array
        Filtered ECG signal.
    rpeaks : array
        R-peak location indices.
    templates_ts : array
        Templates time axis reference (seconds).
    templates : array
        Extracted heartbeat templates.
    heart_rate_ts : array
        Heart rate time axis reference (seconds).
    heart_rate : array
        Instantaneous heart rate (bpm).
    units : str, optional
        Units of the vertical axis. If provided, the plot title will include
        the units information. Default is None.
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show the plot immediately.

    """

    fig = plt.figure(figsize=(10, 5))
    fig.suptitle('ECG Summary', fontsize=12, fontweight='bold')
    gs = gridspec.GridSpec(6, 2)
    fig.subplots_adjust(top=0.88, bottom=0.12, hspace=0.9, wspace=0.3,
                        left=0.1, right=0.96)

    # signal
    ax1 = fig.add_subplot(gs[:2, 0])
    ax1.set_title('Signal')

    ax1.plot(ts, raw, linewidth=MED_LW, label='Raw', alpha=0.6,
             color=color_palette('light-blue'))
    ax1.plot(ts, filtered+np.mean(raw), linewidth=MINOR_LW, label='Filtered',
                color=color_palette('blue'))

    ax1.set_ylabel('Amplitude' if units is None else 'Amplitude (%s)' % units)
    ax1.legend(loc='upper right')
    ax1.tick_params(axis='x', which='both', bottom=False, top=False,
                    labelbottom=False)
    ax1.spines['bottom'].set_visible(False)

    # filtered signal with rpeaks
    ax2 = fig.add_subplot(gs[2:4, 0], sharex=ax1)
    ax2.set_title('R-Peak Detection')

    ymin = np.min(filtered)
    ymax = np.max(filtered)
    alpha = 0.1 * (ymax - ymin)
    ymax += alpha
    ymin -= alpha

    ax2.plot(ts, filtered, linewidth=MINOR_LW, color=color_palette('blue'))
    ax2.plot(ts[rpeaks], filtered[rpeaks], ls='None', marker='x',
             color=color_palette('dark-red'), label='Peaks')

    ax2.set_ylabel('Amplitude' if units is None else 'Amplitude (%s)' % units)
    ax2.legend(loc='upper right')
    ax2.tick_params(axis='x', which='both', bottom=False, top=False,
                    labelbottom=False)
    ax2.spines['bottom'].set_visible(False)

    # heart rate
    ax3 = fig.add_subplot(gs[4:, 0], sharex=ax1)
    ax3.set_title('Heart Rate')

    ax3.plot(heart_rate_ts, heart_rate, linewidth=MAJOR_LW,
             color=color_palette('blue'))

    ax3.set_xlabel('Time (s)')
    ax3.set_ylabel('Heart Rate (bpm)')

    # align y axis labels
    fig.align_ylabels([ax1, ax2, ax3])

    # templates
    ax4 = fig.add_subplot(gs[1:5, 1])
    ax4.set_title('Templates')

    ax4.plot(templates_ts, templates.T, linewidth=MINOR_LW, alpha=0.5,
             color=color_palette('blue'))

    ax4.set_xlabel('Time (s)')
    ax4.set_ylabel('Amplitude' if units is None else 'Amplitude (%s)' % units)

    # add logo
    add_logo(fig)

    # save to file
    if path is not None:
        path = utils.normpath(path)
        root, ext = os.path.splitext(path)
        ext = ext.lower()
        if ext not in ['png', 'jpg']:
            path = root + '.png'

        fig.savefig(path, dpi=200, bbox_inches='tight')

    # show
    if show:
        plt.show()
    else:
        # close
        plt.close(fig)


def plot_bcg(ts=None,
             raw=None,
             filtered=None,
             jpeaks=None,
             templates_ts=None,
             templates=None,
             heart_rate_ts=None,
             heart_rate=None,
             path=None,
             show=False):
    """Create a summary plot from the output of signals.bcg.bcg.

    Parameters
    ----------
    ts : array
        Signal time axis reference (seconds).
    raw : array
        Raw ECG signal.
    filtered : array
        Filtered ECG signal.
    ipeaks : array
        I-peak location indices.
    templates_ts : array
        Templates time axis reference (seconds).
    templates : array
        Extracted heartbeat templates.
    heart_rate_ts : array
        Heart rate time axis reference (seconds).
    heart_rate : array
        Instantaneous heart rate (bpm).
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show the plot immediately.

    """

    fig = plt.figure()
    fig.suptitle('BCG Summary')
    gs = gridspec.GridSpec(6, 2)

    # raw signal
    ax1 = fig.add_subplot(gs[:2, 0])

    ax1.plot(ts, raw, linewidth=MAJOR_LW, label='Raw')

    ax1.set_ylabel('Amplitude')
    ax1.legend()
    ax1.grid()

    # filtered signal with rpeaks
    ax2 = fig.add_subplot(gs[2:4, 0], sharex=ax1)

    ymin = np.min(filtered)
    ymax = np.max(filtered)
    alpha = 0.1 * (ymax - ymin)
    ymax += alpha
    ymin -= alpha

    ax2.plot(ts, filtered, linewidth=MAJOR_LW, label='Filtered')
    ax2.vlines(ts[jpeaks], ymin, ymax,
               color='m',
               linewidth=MINOR_LW,
               label='J-peaks')

    ax2.set_ylabel('Amplitude')
    ax2.legend()
    ax2.grid()

    # heart rate
    ax3 = fig.add_subplot(gs[4:, 0], sharex=ax1)

    ax3.plot(heart_rate_ts, heart_rate, linewidth=MAJOR_LW, label='Heart Rate')

    ax3.set_xlabel('Time (s)')
    ax3.set_ylabel('Heart Rate (bpm)')
    ax3.legend()
    ax3.grid()

    # templates
    ax4 = fig.add_subplot(gs[1:5, 1])

    ax4.plot(templates_ts, templates.T, 'm', linewidth=MINOR_LW, alpha=0.7)

    ax4.set_xlabel('Time (s)')
    ax4.set_ylabel('Amplitude')
    ax4.set_title('Templates')
    ax4.grid()

    # make layout tight
    gs.tight_layout(fig)

    # save to file
    if path is not None:
        path = utils.normpath(path)
        root, ext = os.path.splitext(path)
        ext = ext.lower()
        if ext not in ['png', 'jpg']:
            path = root + '.png'

        fig.savefig(path, dpi=200, bbox_inches='tight')

    # show
    if show:
        plt.show()
    else:
        # close
        plt.close(fig)


def plot_pcg(ts=None,
             raw=None,
             filtered=None,
             peaks=None,
             heart_sounds=None,
             heart_rate_ts=None,
             inst_heart_rate=None,
             units=None,
             path=None,
             show=False):
    """Create a summary plot from the output of signals.pcg.pcg.
    Parameters
    ----------
    ts : array
        Signal time axis reference (seconds).
    raw : array
        Raw PCG signal.
    filtered : array
        Filtered PCG signal.
    peaks : array
        Peak location indices.
    heart_sounds : array
        Classification of peaks as S1 or S2
    heart_rate_ts : array
        Heart rate time axis reference (seconds).
    inst_heart_rate : array
        Instantaneous heart rate (bpm).
    units : str, optional
        Units of the vertical axis. If provided, the plot title will include
        the units information. Default is None.
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show the plot immediately.

    """

    fig = plt.figure(figsize=(12, 5))
    fig.suptitle('PCG Summary', fontsize=12, fontweight='bold')
    gs = gridspec.GridSpec(6, 2)
    fig.subplots_adjust(top=0.88, bottom=0.12, hspace=0.9, wspace=0.3,
                        left=0.1, right=0.96)

    # signal
    ax1 = fig.add_subplot(gs[:2, 0])
    ax1.set_title('Signal')

    ax1.plot(ts, raw, linewidth=MED_LW, label='Raw',
             color=color_palette('light-blue'))
    ax1.plot(ts, filtered+np.mean(raw), linewidth=MINOR_LW, label='Filtered',
                color=color_palette('blue'))

    ax1.set_ylabel('Amplitude' if units is None else 'Amplitude (%s)' % units)
    ax1.legend(loc='upper right')
    ax1.tick_params(axis='x', which='both', bottom=False, top=False,
                    labelbottom=False)
    ax1.spines['bottom'].set_visible(False)

    # filtered signal with sounds
    ax2 = fig.add_subplot(gs[2:4, 0], sharex=ax1)
    ax2.set_title('Heart Sound Detection')

    ymin = np.min(filtered)
    ymax = np.max(filtered)
    alpha = 0.1 * (ymax - ymin)
    ymax += alpha
    ymin -= alpha

    ax2.plot(ts, filtered, linewidth=MINOR_LW, color=color_palette('blue'))
    ax2.vlines(ts[peaks], ymin, ymax,
               color=color_palette('dark-red'),
               linewidth=MED_LW,
               label='Heart Sounds')

    ax2.set_ylabel('Amplitude' if units is None else 'Amplitude (%s)' % units)
    ax2.legend(loc='upper right')
    ax2.tick_params(axis='x', which='both', bottom=False, top=False,
                    labelbottom=False)
    ax2.spines['bottom'].set_visible(False)

    # heart rate
    ax3 = fig.add_subplot(gs[4:, 0], sharex=ax1)
    ax3.set_title('Heart Rate')

    ax3.plot(heart_rate_ts, inst_heart_rate, linewidth=MAJOR_LW,
             color=color_palette('blue'))

    ax3.set_xlabel('Time (s)')
    ax3.set_ylabel('Heart Rate (bpm)')

    # align y axis labels
    fig.align_ylabels([ax1, ax2, ax3])

    # heart sounds
    ax4 = fig.add_subplot(gs[1:5, 1], sharex=ax1)

    ax4.plot(ts, filtered, linewidth=MINOR_LW, color=color_palette('blue'))
    for i in range(0, len(peaks)):
        text = "S" + str(int(heart_sounds[i]))
        plt.annotate(text, (ts[peaks[i]], ymax - alpha), ha='center', va='center', size=9,
                     color=color_palette('dark-grey'))

    ax4.set_xlabel('Time (s)')
    ax4.set_ylabel('Amplitude' if units is None else 'Amplitude (%s)' % units)
    ax4.set_title('Heart Sounds Classification')

    # add logo
    add_logo(fig)

    # save to file
    if path is not None:
        path = utils.normpath(path)
        root, ext = os.path.splitext(path)
        ext = ext.lower()
        if ext not in ['png', 'jpg']:
            path = root + '.png'

        fig.savefig(path, dpi=200, bbox_inches='tight')

    # show
    if show:
        plt.show()
    else:
        # close
        plt.close(fig)


def _plot_rates(thresholds, rates, variables,
                lw=1,
                colors=None,
                alpha=1,
                eer_idx=None,
                labels=False,
                ax=None):
    """Plot biometric rates.

    Parameters
    ----------
    thresholds : array
        Classifier thresholds.
    rates : dict
        Dictionary of rates.
    variables : list
        Keys from 'rates' to plot.
    lw : int, float, optional
        Plot linewidth.
    colors : list, optional
        Plot line color for each variable.
    alpha : float, optional
        Plot line alpha value.
    eer_idx : int, optional
        Classifier reference index for the Equal Error Rate.
    labels : bool, optional
        If True, will show plot labels.
    ax : axis, optional
        Plot Axis to use.

    Returns
    -------
    fig : Figure
        Figure object.

    """

    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    else:
        fig = ax.figure

    if colors is None:
        x = np.linspace(0., 1., len(variables))
        colors = plt.get_cmap('rainbow')(x)

    if labels:
        for i, v in enumerate(variables):
            ax.plot(thresholds, rates[v], colors[i],
                    lw=lw,
                    alpha=alpha,
                    label=v)
    else:
        for i, v in enumerate(variables):
            ax.plot(thresholds, rates[v], colors[i], lw=lw, alpha=alpha)

    if eer_idx is not None:
        x, y = rates['EER'][eer_idx]
        ax.vlines(x, 0, 1, 'r', lw=lw)
        ax.set_title('EER = %0.2f %%' % (100. * y))

    return fig


def plot_biometrics(assessment=None, eer_idx=None, path=None, show=False):
    """Create a summary plot of a biometrics test run.

    Parameters
    ----------
    assessment : dict
        Classification assessment results.
    eer_idx : int, optional
        Classifier reference index for the Equal Error Rate.
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show the plot immediately.

    """

    fig = plt.figure()
    fig.suptitle('Biometrics Summary')

    c_sub = ['#008bff', '#8dd000']
    c_global = ['#0037ff', 'g']

    ths = assessment['thresholds']

    auth_ax = fig.add_subplot(121)
    id_ax = fig.add_subplot(122)

    # subject results
    for sub in six.iterkeys(assessment['subject']):
        auth_rates = assessment['subject'][sub]['authentication']['rates']
        _ = _plot_rates(ths, auth_rates, ['FAR', 'FRR'],
                        lw=MINOR_LW,
                        colors=c_sub,
                        alpha=0.4,
                        eer_idx=None,
                        labels=False,
                        ax=auth_ax)

        id_rates = assessment['subject'][sub]['identification']['rates']
        _ = _plot_rates(ths, id_rates, ['MR', 'RR'],
                        lw=MINOR_LW,
                        colors=c_sub,
                        alpha=0.4,
                        eer_idx=None,
                        labels=False,
                        ax=id_ax)

    # global results
    auth_rates = assessment['global']['authentication']['rates']
    _ = _plot_rates(ths, auth_rates, ['FAR', 'FRR'],
                    lw=MAJOR_LW,
                    colors=c_global,
                    alpha=1,
                    eer_idx=eer_idx,
                    labels=True,
                    ax=auth_ax)

    id_rates = assessment['global']['identification']['rates']
    _ = _plot_rates(ths, id_rates, ['MR', 'RR'],
                    lw=MAJOR_LW,
                    colors=c_global,
                    alpha=1,
                    eer_idx=eer_idx,
                    labels=True,
                    ax=id_ax)

    # set labels and grids
    auth_ax.set_xlabel('Threshold')
    auth_ax.set_ylabel('Authentication')
    auth_ax.grid()
    auth_ax.legend()

    id_ax.set_xlabel('Threshold')
    id_ax.set_ylabel('Identification')
    id_ax.grid()
    id_ax.legend()

    # make layout tight
    fig.tight_layout()

    # save to file
    if path is not None:
        path = utils.normpath(path)
        root, ext = os.path.splitext(path)
        ext = ext.lower()
        if ext not in ['png', 'jpg']:
            path = root + '.png'

        fig.savefig(path, dpi=200, bbox_inches='tight')

    # show
    if show:
        plt.show()
    else:
        # close
        plt.close(fig)


def plot_clustering(data=None, clusters=None, path=None, show=False):
    """Create a summary plot of a data clustering.

    Parameters
    ----------
    data : array
        An m by n array of m data samples in an n-dimensional space.
    clusters : dict
        Dictionary with the sample indices (rows from `data`) for each cluster.
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show the plot immediately.

    """

    fig = plt.figure()
    fig.suptitle('Clustering Summary')

    ymin, ymax = _yscaling(data, alpha=1.2)

    # determine number of clusters
    keys = list(clusters)
    nc = len(keys)

    if nc <= 4:
        nrows = 2
        ncols = 4
    else:
        area = nc + 4

        # try to fit to a square
        nrows = int(np.ceil(np.sqrt(area)))

        if nrows > MAX_ROWS:
            # prefer to increase number of columns
            nrows = MAX_ROWS

        ncols = int(np.ceil(area / float(nrows)))

    # plot grid
    gs = gridspec.GridSpec(nrows, ncols, hspace=0.2, wspace=0.2)

    # global axes
    ax_global = fig.add_subplot(gs[:2, :2])

    # cluster axes
    c_grid = np.ones((nrows, ncols), dtype='bool')
    c_grid[:2, :2] = False
    c_rows, c_cols = np.nonzero(c_grid)

    # generate color map
    x = np.linspace(0., 1., nc)
    cmap = plt.get_cmap('rainbow')

    for i, k in enumerate(keys):
        aux = data[clusters[k]]
        color = cmap(x[i])
        label = 'Cluster %s' % k
        ax = fig.add_subplot(gs[c_rows[i], c_cols[i]], sharex=ax_global)
        ax.set_ylim([ymin, ymax])
        ax.set_title(label)
        ax.grid()

        if len(aux) > 0:
            ax_global.plot(aux.T, color=color, lw=MINOR_LW, alpha=0.7)
            ax.plot(aux.T, color=color, lw=MAJOR_LW)

    ax_global.set_title('All Clusters')
    ax_global.set_ylim([ymin, ymax])
    ax_global.grid()

    # make layout tight
    gs.tight_layout(fig)

    # save to file
    if path is not None:
        path = utils.normpath(path)
        root, ext = os.path.splitext(path)
        ext = ext.lower()
        if ext not in ['png', 'jpg']:
            path = root + '.png'

        fig.savefig(path, dpi=200, bbox_inches='tight')

    # show
    if show:
        plt.show()
    else:
        # close
        plt.close(fig)


def plot_rri(rri, rri_trend=None, legends=None, ax=None, show=False):
    """Plot a series of RR intervals.

    Parameters
    ----------
    rri : array
        RR-intervals (ms).
    rri_trend : array, optional
        RR-intervals trend (ms).
    legends : dict, optional
        Dictionary of features to add to the plot legend.
    ax : axis, optional
        Plot Axis to use.
    show : bool, optional
        If True, show the plot immediately.
    """

    # time axis
    t = np.cumsum(rri) / 1000.

    # plot
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(8, 4))
        fig.suptitle('HRV - RR Intervals', fontsize=12, fontweight='bold')
        fig.subplots_adjust(bottom=0.17)
        add_logo(fig)

    # plot signal
    ax.plot(t, rri, color=color_palette('blue'), linewidth=MAJOR_LW,
            label='RR Intervals')
    ax.set_ylabel('RRI (ms)')
    ax.set_xlabel('Time (s)')
    if rri_trend is not None:
        ax.plot(t, rri_trend, color=color_palette('dark-red'), alpha=0.5,
                linewidth=MED_LW, label='Trend')

    # plot legend
    if legends is not None:
        handles, labels = ax.get_legend_handles_labels()
        for key, value in legends.items():
            new_patch = patches.Patch(color='white', alpha=0)
            handles.extend([new_patch])
            labels.extend(['%s = %.2f %s' % (key, value[0], value[1])])

        try:
            pos = ax.get_position()
            ax.set_position([pos.x0, pos.y0, pos.width * 0.9, pos.height])
            ax.legend(handles=handles, labels=labels, loc='upper left',
                      bbox_to_anchor=(1.01, 1.02), frameon=False)
        except ValueError:
            pass
    else:
        ax.legend(loc='upper right')

    # show
    if show:
        plt.show()


def plot_poincare(rri=None,
                  s=None,
                  sd1=None,
                  sd2=None,
                  legends=None,
                  ax=None,
                  show=False):
    """Plot a Poincaré plot of a series of RR intervals (RRI[i+1] vs. RRI[i])
    from the output of signals.hrv.compute_poincare.

    Parameters
    ----------
    rri : array
        RR-intervals (ms).
    s : float
        S - Area of the ellipse of the Poincaré plot (ms^2).
    sd1 : float
        SD1 - Poincaré plot standard deviation perpendicular to the identity
        line (ms).
    sd2 : float
        SD2 - Poincaré plot standard deviation along the identity line (ms).
    legends : dict, optional
        Dictionary of features to add to the plot legend.
    ax : axis, optional
        Plot Axis to use.
    show : bool, optional
        If True, show the plot immediately.
    """

    x, y = rri[:-1], rri[1:]
    rr_mean = rri.mean()

    if ax is None:
        fig, ax = plt.subplots(figsize=(7, 5))
        fig.suptitle('HRV - Poincaré Plot', fontsize=12, fontweight='bold')
        # add logo
        add_logo(fig)

    ax.set_xlabel('$RR_i$ (ms)')
    ax.set_ylabel('$RR_{i+1}$ (ms)')

    # plot Poincaré data points
    ax.scatter(x, y, marker='.', color=color_palette('blue'), alpha=0.5, s=100,
               zorder=1)
    ax.set_xlim([np.min(rri) - 50, np.max(rri) + 50])
    ax.set_ylim([np.min(rri) - 50, np.max(rri) + 50])
    ax.set_aspect(1. / ax.get_data_ratio())

    # draw identity line (RRi+1=RRi)
    lims = [np.min([ax.get_xlim(), ax.get_ylim()]),  # min of both axes
            np.max([ax.get_xlim(), ax.get_ylim()])]  # max of both axes

    ax.plot(lims, lims, linewidth=0.7, color=color_palette('grey'),
            linestyle='--', zorder=2, label='Identity line')

    # draw ellipse
    ellipse = patches.Ellipse((rr_mean, rr_mean), sd1 * 2, sd2 * 2, angle=-45,
                              linewidth=1, edgecolor=color_palette('dark-grey'),
                              facecolor='None', label='S = %.1f ms$^2$' % s, zorder=3)
    ax.add_artist(ellipse)

    # draw SD1 and SD2
    ax.arrow(rr_mean, rr_mean, sd1 * np.cos(3 * np.pi / 4), sd1 * np.sin(3 * np.pi / 4),
             facecolor=color_palette('dark-red'), edgecolor=color_palette('dark-red'),
             linewidth=2, length_includes_head=True, head_width=4, head_length=4,
             label='SD1 = %.1f ms' % sd1, zorder=3)

    ax.arrow(rr_mean, rr_mean, sd2 * np.cos(np.pi / 4), sd2 * np.sin(np.pi / 4),
             facecolor=color_palette('yellow'), edgecolor=color_palette('yellow'),
             linewidth=2, length_includes_head=True, head_width=4, head_length=4,
             label='SD2 = %.1f ms' % sd2, zorder=3)

    # draw SD1 and SD2 axes
    f = 1.25  # scaling factor
    ax.add_artist(
        lines.Line2D([rr_mean - f * sd1 * np.cos(3 * np.pi / 4), rr_mean + f * sd1 * np.cos(3 * np.pi / 4)],
                     [rr_mean + f * sd1 * np.cos(3 * np.pi / 4), rr_mean - f * sd1 * np.cos(3 * np.pi / 4)],
                     lw=1, color='0.2'))

    ax.add_artist(lines.Line2D([rr_mean - f * sd2 * np.cos(np.pi / 4), rr_mean + f * sd2 * np.cos(np.pi / 4)],
                               [rr_mean - f * sd2 * np.cos(np.pi / 4), rr_mean + f * sd2 * np.cos(np.pi / 4)],
                               lw=1, color='0.2'))
    # adjust grid
    ax.set_axisbelow(True)

    # add extra labels
    # update figure legend
    handles, labels = ax.get_legend_handles_labels()
    if legends is not None:
        for key, value in legends.items():
            if value is not None:
                new_patch = patches.Patch(color='white', alpha=0)
                handles.extend([new_patch])
                labels.extend(['%s = %.2f' % (key, value)])

    # change handle for ellipse and arrows
    handles[1] = plt.Line2D([], [], color=color_palette('dark-grey'),
                            marker="o", markersize=10, linewidth=0,
                            markerfacecolor='none')  # ellipse
    handles[2] = plt.Line2D([], [], color=color_palette('dark-red'),
                            linestyle="-", linewidth=1)  # SD1
    handles[3] = plt.Line2D([], [], color=color_palette('yellow'),
                            linestyle="-", linewidth=1)  # SD2

    # create legend
    pos = ax.get_position()
    ax.set_position([pos.x0, pos.y0, pos.width * 0.9, pos.height])
    ax.legend(handles=handles, labels=labels, loc='upper left',
              bbox_to_anchor=(1.01, 1.02), frameon=False)

    # show
    if show:
        plt.show()


def plot_hrv_hist(rri=None,
                  bins=None,
                  q_hist=None,
                  hti=None,
                  tinn=None,
                  ax=None,
                  show=False):
    """Plots the RRI histogram with the corresponding geometrical HRV features
    from the output of signals.hrv.compute_geometrical.

    Parameters
    ----------
    rri : array
        RR-intervals (ms).
    bins : array
        Histogram bins.
    q_hist : array
        Multilinear function fitted to the histogram.
    hti : float
        HTI - HRV triangular index - Integral of the density of the RR interval
        histogram divided by its height.
    tinn : float
        TINN - Baseline width of RR interval histogram (ms).
    ax : axis, optional
        Plot Axis to use.
    show : bool, optional
        If True, show the plot immediately.
    """

    # plot histogram and triangle
    if ax is None:
        fig, ax = plt.subplots()
        fig.suptitle('HRV - RRI Distribution', fontsize=12, fontweight='bold')

        # add logo
        add_logo(fig)

    ax.hist(rri, bins, facecolor=color_palette('light-blue'),
            edgecolor=color_palette('dark-grey'), label='HTI: %.1f' % hti)
    ax.plot(bins, q_hist, color=color_palette('dark-red'), linewidth=1.5,
            label='TINN: %.1f ms' % tinn)
    ax.set_xlabel('RR Interval (ms)')
    ax.set_ylabel('Count')
    ax.locator_params(axis='y', integer=True)
    pos = ax.get_position()
    ax.set_position([pos.x0, pos.y0, pos.width * 0.9, pos.height])
    ax.legend(loc='upper left',
              bbox_to_anchor=(1.01, 1.02), frameon=False)

    # adjust grid to be in the background
    ax.set_axisbelow(True)

    # show
    if show:
        plt.show()


def plot_hrv_fbands(frequencies=None,
                    powers=None,
                    fbands=None,
                    method_name=None,
                    legends=None,
                    ax=None,
                    show=False):
    """Plots the power spectrum and highlights the defined frequency bands
    from the output of signals.hrv.compute_fbands.

    Parameters
    ----------
    frequencies : array
        Frequency axis.
    powers : array
        Power spectrum values for the frequency axis.
    fbands : dict, optional
        Dictionary containing the limits of the frequency bands.
    method_name : str, optional
        Method that was used to compute the power spectrum.
    legends : dict, optional
        Additional legend elements.
    ax : axis, optional
        Plot Axis to use.
    show : bool, optional
        If True, show the plot immediately.
    """

    spectrum_colors = {'ulf': '#e6eff6',
                       'vlf': '#89b4c4',
                       'lf': '#548999',
                       'hf': '#f1d3a1',
                       'vhf': '#e3dbd9'
                       }

    # convert power values
    powers = powers / (10 ** 6)  # to s^2/Hz

    # initialize plot
    if ax is None:
        fig, ax = plt.subplots()
        fig.suptitle('HRV - Power Spectral Density', fontsize=12, fontweight='bold')
        # add logo
        add_logo(fig)

    # figure attributes
    if method_name is not None:
        ax.set_title(f'({method_name})')
    ax.set_xlabel('Frequency (Hz)')
    ax.set_ylabel('Power (s$^2$/Hz)')
    ax.set_xlim([0, 0.5])

    # plot spectrum
    ax.plot(frequencies, powers, linewidth=1, color='0.2')
    ax.margins(0)

    # plot frequency bands
    for fband in fbands.keys():
        band = np.argwhere((frequencies >= fbands[fband][0]) & (frequencies <= fbands[fband][-1])).reshape(-1)
        color = spectrum_colors[fband]
        if len(band) > 0:
            ax.fill_between(frequencies[band], powers[band], color=color,
                            label=f'{fband.upper()}: {legends[fband+"_rpwr"]*100:.1f}%')

    # update figure legend
    handles, labels = ax.get_legend_handles_labels()
    if legends.__len__() != 0:
        for key, value in legends.items():
            if not key.endswith('_rpwr') and value is not None:
                new_patch = patches.Patch(color='white', alpha=0)
                handles.extend([new_patch])
                labels.extend(['%s = %.2f' % (key, value)])
    pos = ax.get_position()
    ax.set_position([pos.x0, pos.y0, pos.width * 0.9, pos.height])
    ax.legend(handles=handles, labels=labels, loc='upper left',
              bbox_to_anchor=(1.01, 1.02), frameon=False)

    # adjust grid
    ax.set_axisbelow(True)

    # show
    if show:
        plt.show()


def plot_hrv(rri,
             rri_trend=None,
             td_out=None,
             nl_out=None,
             fd_out=None,
             show=False):
    """Create a summary plot of a HRV analysis.

    Parameters
    ----------
    rri : array
        RR-intervals (ms).
    rri_trend : array, optional
        RR-intervals trend (ms).
    td_out : dict
        Output of signals.hrv.timedomain.
    nl_out : dict
        Output of signals.hrv.nonlinear.
    fd_out : dict
        Output of signals.hrv.frequencyomain.
    show : bool, optional
        If True, show the plot immediately.
    """

    # convert ReturnTuple to dict
    if td_out is not None:
        td_out = td_out.as_dict()
    if nl_out is not None:
        nl_out = nl_out.as_dict()
    if fd_out is not None:
        fd_out = fd_out.as_dict()

    # plot
    fig = plt.figure(figsize=(12, 6))
    fig.suptitle('HRV Summary', fontsize=12, fontweight='bold')
    gs = gridspec.GridSpec(6, 2)

    # plot rri and display time domain features
    ax1 = fig.add_subplot(gs[:2, 0])
    time_legends = {'Mean': (td_out['rr_mean'], 'ms'),
                    'Median': (td_out['rr_median'], 'ms'),
                    'MinMax': (td_out['rr_minmax'], 'ms'),
                    'SDNN': (td_out['sdnn'], 'ms'),
                    'RMSSD': (td_out['rmssd'], 'ms'),
                    'pNN50': (td_out['pnn50'], '%'),
                    }
    plot_rri(rri=rri, rri_trend=rri_trend, legends=time_legends, ax=ax1,
             show=False)
    ax1.set_title('Time Domain', fontweight='bold')

    # plot hrv hist
    ax2 = fig.add_subplot(gs[2:, 0])
    bins = td_out['bins']
    q_hist = td_out['q_hist']
    hti = td_out['hti']
    tinn = td_out['tinn']
    plot_hrv_hist(rri=rri, bins=bins, q_hist=q_hist, hti=hti, tinn=tinn,
                  ax=ax2, show=False)

    # plot pointcare and display non-linear features
    ax3 = fig.add_subplot(gs[:3, 1])
    ax3.set_title('Non-linear Domain', fontweight='bold')
    rri_pc = rri - rri_trend if rri_trend is not None else rri
    s = nl_out['s']
    sd1 = nl_out['sd1']
    sd2 = nl_out['sd2']
    poincare_legend = {'SD1/SD2': nl_out['sd12'],
                       'SD2/SD1': nl_out['sd21'],
                       'SampEn': nl_out['sampen'] if 'sampen' in nl_out.keys() else None,
                       'AppEn': nl_out['appen'] if 'appen' in nl_out.keys() else None,
                       }
    plot_poincare(rri=rri_pc, s=s, sd1=sd1, sd2=sd2, legends=poincare_legend,
                  ax=ax3, show=False)

    # plot hrv fbands
    ax4 = fig.add_subplot(gs[3:, 1])
    frequencies = fd_out['frequencies']
    powers = fd_out['powers']
    fbands = fd_out['fbands']
    method_name = fd_out['freq_method']
    freq_legends = {'LF/HF': fd_out['lf_hf']}
    for key in fd_out.keys():
        if key.endswith('_rpwr'):
            freq_legends[key] = fd_out[key]
    plot_hrv_fbands(frequencies=frequencies, powers=powers, fbands=fbands,
                    method_name=method_name, legends=freq_legends, ax=ax4,
                    show=False)
    ax4.set_title('Frequency Domain', fontweight='bold')

    # tight layout
    gs.tight_layout(fig)

    # adjust margins
    gs.update(right=0.88, bottom=0.14)

    # add logo
    add_logo(fig)

    # show
    if show:
        plt.show()



# -*- coding: utf-8 -*-
"""
biosppy.features.time_freq
--------------------------

This module provides methods to extract time-frequency features using
discrete wavelet decomposition.

:copyright: (c) 2015-2023 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# Imports
# 3rd party
import numpy as np
import pywt

# local
from .. import utils
from ..signals import tools as st


def time_freq(signal=None, wavelet="db4", level=5):
    """Compute statistical metrics over the signal discrete wavelet transform
    approximation and detail coefficients.

    Parameters
    ----------
    signal : array
        Input signal.
    wavelet: str
        Type of wavelet. Default is db4 (Daubechies 4).
    level: int
        Decomposition level. Default is 5.

    Returns
    -------
    dwt_app_{metric} : float
        Statistical metrics over the approximation coefficients.
    dwt_det{level}_{metric} : float
        Statistical metrics over the detail coefficients at the specified level.

    Notes
    -----
    Check biosppy.signals.tools.signal_stats for the list of available metrics.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    # initialize output
    feats = utils.ReturnTuple((), ())

    # compute wavelet
    coeffs = compute_wavelet(signal=signal, wavelet=wavelet, level=level)

    # compute stats
    for coeff, coeff_name in zip(coeffs, coeffs.keys()):
        signal_feats = st.signal_stats(coeff)
        for stat_value, stat_name in zip(signal_feats, signal_feats.keys()):
            feats = feats.append(stat_value, coeff_name + '_' + stat_name)

    return feats


def compute_wavelet(signal=None, wavelet="db4", level=5):
    """Compute the approximation and highest detail coefficients of the signal
    using the discrete wavelet transform.

    Parameters
    ----------
    signal : array
        Input signal.
    wavelet: str
        Type of wavelet.
    level: int
        Decomposition level

    Returns
    -------
    dwt_app : array
        Approximation coefficients.
    dwt_det{level} : array
        Detail coefficients at the specified level.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    # wavelets
    dwt_det = pywt.downcoef("d", signal, wavelet, level)
    dwt_app = pywt.downcoef("a", signal, wavelet, level)

    # output
    args = (dwt_app, dwt_det)
    names = ("dwt_app", f"dwt_det{level}")

    return utils.ReturnTuple(args, names)



# -*- coding: utf-8 -*-
"""
biosppy.quality
-------------

This provides functions to assess the quality of several biosignals.

:copyright: (c) 2015-2023 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# Imports
# compat
from __future__ import absolute_import, division, print_function

# local
from . import utils
from .signals import ecg, tools

# 3rd party
import numpy as np


def quality_eda(x=None, methods=['bottcher'], sampling_rate=None):
    """Compute the quality index for one EDA segment.

        Parameters
        ----------
        x : array
            Input signal to test.
        methods : list
            Method to assess quality. One or more of the following: 'bottcher'.
        sampling_rate : int
            Sampling frequency (Hz).
        Returns
        -------
        args : tuple
            Tuple containing the quality index for each method.
        names : tuple
            Tuple containing the name of each method.
        """
    # check inputs
    if x is None:
        raise TypeError("Please specify the input signal.")
    
    if sampling_rate is None:
        raise TypeError("Please specify the sampling rate.")
    
    assert len(x) > sampling_rate * 2, 'Segment must be 5s long'

    args, names = (), ()
    available_methods = ['bottcher']

    for method in methods:

        assert method in available_methods, "Method should be one of the following: " + ", ".join(available_methods)
    
        if method == 'bottcher':
            quality = eda_sqi_bottcher(x, sampling_rate)
    
        args += (quality,)
        names += (method,)

    return utils.ReturnTuple(args, names)


def quality_ecg(segment, methods=['Level3'], sampling_rate=None, 
                fisher=True, f_thr=0.01, threshold=0.9, bit=0, 
                nseg=1024, num_spectrum=[5, 20], dem_spectrum=None, 
                mode_fsqi='simple'):
    
    """Compute the quality index for one ECG segment.

    Parameters
    ----------
    segment : array
        Input signal to test.
    method : string
        Method to assess quality. One of the following: 'Level3', 'pSQI', 'kSQI', 'fSQI'.
    sampling_rate : int
        Sampling frequency (Hz).
    threshold : float
        Threshold for the correlation coefficient.
    bit : int
        Number of bits of the ADC. Resolution bits, for the BITalino is 10 bits.

    Returns
    -------
    args : tuple
        Tuple containing the quality index for each method.
    names : tuple
        Tuple containing the name of each method.
    """
    args, names = (), ()
    available_methods = ['Level3', 'pSQI', 'kSQI', 'fSQI']

    for method in methods:

        assert method in available_methods, 'Method should be one of the following: ' + ', '.join(available_methods)

        if method == 'Level3':
            # returns a SQI level 0, 0.5 or 1.0
            quality = ecg_sqi_level3(segment, sampling_rate, threshold, bit)

        elif method == 'pSQI':
            quality = ecg.pSQI(segment, f_thr=f_thr)
        
        elif method == 'kSQI':
            quality = ecg.kSQI(segment, fisher=fisher)

        elif method == 'fSQI':
            quality = ecg.fSQI(segment, fs=sampling_rate, nseg=nseg, num_spectrum=num_spectrum, dem_spectrum=dem_spectrum, mode=mode_fsqi)

        args += (quality,)
        names += (method,)

    return utils.ReturnTuple(args, names)


def ecg_sqi_level3(segment, sampling_rate, threshold, bit):

    """Compute the quality index for one ECG segment. The segment should have 10 seconds.


    Parameters
    ----------
    segment : array
        Input signal to test.
    sampling_rate : int
        Sampling frequency (Hz).
    threshold : float
        Threshold for the correlation coefficient.
    bit : int
        Number of bits of the ADC.? Resolution bits, for the BITalino is 10 bits.
    
    Returns
    -------
    quality : string
        Signal Quality Index ranging between 0 (LQ), 0.5 (MQ) and 1.0 (HQ).

    """
    LQ, MQ, HQ = 0.0, 0.5, 1.0
    
    if bit !=  0:
        if (max(segment) - min(segment)) >= (2**bit - 1):
            return LQ
    if sampling_rate is None:
        raise IOError('Sampling frequency is required')
    if len(segment) < sampling_rate * 5:
        raise IOError('Segment must be 5s long')
    else:
        # TODO: compute ecg quality when in contact with the body
        rpeak1 = ecg.hamilton_segmenter(segment, sampling_rate=sampling_rate)['rpeaks']
        rpeak1 = ecg.correct_rpeaks(signal=segment, rpeaks=rpeak1, sampling_rate=sampling_rate, tol=0.05)['rpeaks']
        if len(rpeak1) < 2:
            return LQ
        else:
            hr = sampling_rate * (60/np.diff(rpeak1))
            quality = MQ if (max(hr) <= 200 and min(hr) >= 40) else LQ
        if quality == MQ:
            templates, _ = ecg.extract_heartbeats(signal=segment, rpeaks=rpeak1, sampling_rate=sampling_rate, before=0.2, after=0.4)
            corr_points = np.corrcoef(templates)
            if np.mean(corr_points) > threshold:
                quality = HQ

    return quality 


def eda_sqi_bottcher(x=None, sampling_rate=None):  # -> Timeline
    """
    Suggested by Böttcher et al. Scientific Reports, 2022, for wearable wrist EDA.

    This is given by a binary score 0/1 defined by the following rules:
     - mean of the segment of 2 seconds should be > 0.05
     - rate of amplitude change (given by racSQI) should be < 0.2
    This score is calculated for each 2 seconds window of the segment. The average of the scores is the final SQI.

    This method was designed for a segment of 60s

    """
    quality_score = 0
    if x is None:
        raise TypeError("Please specify the input signal.")
    if sampling_rate is None:
        raise TypeError("Please specify the sampling rate.")
    
    if len(x) < sampling_rate * 60:
        print("This method was designed for a signal of 60s but will be applied to a signal of {}s".format(len(x)/sampling_rate))
    # create segments of 2 seconds
    segments_2s = x.reshape(-1, int(sampling_rate*2))
    ## compute racSQI for each segment
    # first compute the min and max of each segment
    min_ = np.min(segments_2s, axis=1)
    max_ = np.max(segments_2s, axis=1)
    # then compute the RAC (max-min)/max
    rac = np.abs((max_ - min_) / max_)
    # ratio will be 1 if the rac is < 0.2 and if the mean of the segment is > 0.05 and will be 0 otherwise
    quality_score = ((rac < 0.2) & (np.mean(segments_2s, axis=1) > 0.05)).astype(int)
    # the final SQI is the average of the scores 
    return np.mean(quality_score)
    



# -*- coding: utf-8 -*-
"""
biosppy.stats
-------------

This module provides statistical functions and related tools.

:copyright: (c) 2015-2023 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# Imports
# compat
from __future__ import absolute_import, division, print_function
import six

# local
from . import utils
from .signals import tools

# 3rd party
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import pearsonr, ttest_rel, ttest_ind


def pearson_correlation(x=None, y=None):
    """Compute the Pearson Correlation Coefficient between two signals.

    The coefficient is given by:

    .. math::

        r_{xy} = \\frac{E[(X - \\mu_X) (Y - \\mu_Y)]}{\\sigma_X \\sigma_Y}

    Parameters
    ----------
    x : array
        First input signal.
    y : array
        Second input signal.

    Returns
    -------
    r : float
        Pearson correlation coefficient, ranging between -1 and +1.
    pvalue : float
        Two-tailed p-value. The p-value roughly indicates the probability of
        an uncorrelated system producing datasets that have a Pearson correlation
        at least as extreme as the one computed from these datasets.

    Raises
    ------
    ValueError
        If the input signals do not have the same length.

    """

    # check inputs
    if x is None:
        raise TypeError("Please specify the first input signal.")

    if y is None:
        raise TypeError("Please specify the second input signal.")

    # ensure numpy
    x = np.array(x)
    y = np.array(y)

    n = len(x)

    if n != len(y):
        raise ValueError("Input signals must have the same length.")

    r, pvalue = pearsonr(x, y)

    args = (r, pvalue)
    names = ('r', 'pvalue')

    return utils.ReturnTuple(args, names)


def linear_regression(x=None, y=None, show=True):
    """Plot the linear regression between two signals and get the equation coefficients.

    The linear regression uses the least squares method.

    Parameters
    ----------
    x : array
        First input signal.
    y : array
        Second input signal.
    show : bool
        If True, show the plot.

    Returns
    -------
    coeffs : array
        Linear regression coefficients: [m, b].

    Raises
    ------
    ValueError
        If the input signals do not have the same length.

    """

    # check inputs
    if x is None:
        raise TypeError("Please specify the first input signal.")

    if y is None:
        raise TypeError("Please specify the second input signal.")

    # ensure numpy
    x = np.array(x)
    y = np.array(y)

    n = len(x)

    if n != len(y):
        raise ValueError("Input signals must have the same length.")

    coeffs = np.polyfit(x, y, 1)
    f = np.poly1d(coeffs)

    x_min = x.min()
    x_max = x.max()

    y_min = f(x_min)
    y_max = f(x_max)

    if show:
        plt.scatter(x, y)
        plt.plot(
            [x_min, x_max],
            [y_min, y_max],
            c="orange",
            label="y={:.3f}x+{:.3f}".format(coeffs[0], coeffs[1]),
        )
        plt.title("Linear Regression")
        plt.xlabel("x")
        plt.ylabel("y")
        plt.legend()

    args = (coeffs[0], coeffs[1])
    labels = ["m", "b"]

    return utils.ReturnTuple(args, labels)


def paired_test(x=None, y=None):
    """
    Perform the Student's paired t-test on the arrays x and y.
    This is a two-sided test for the null hypothesis that 2 related
    or repeated samples have identical average (expected) values.

    Parameters
    ----------
    x : array
        First input signal.
    y : array
        Second input signal.

    Returns
    -------
    statistic : float
        t-statistic. The t-statistic is used in a t-test to determine
        if you should support or reject the null hypothesis.
    pvalue : float
        Two-sided p-value.

    Raises
    ------
    ValueError
        If the input signals do not have the same length.

    """

    # check inputs
    if x is None:
        raise TypeError("Please specify the first input signal.")

    if y is None:
        raise TypeError("Please specify the second input signal.")

    # ensure numpy
    x = np.array(x)
    y = np.array(y)

    n = len(x)

    if n != len(y):
        raise ValueError("Input signals must have the same length.")

    statistic, pvalue = ttest_rel(x, y)

    return statistic, pvalue


def unpaired_test(x=None, y=None):
    """
    Perform the Student's unpaired t-test on the arrays x and y.
    This is a two-sided test for the null hypothesis that 2 independent
    samples have identical average (expected) values. This test assumes
    that the populations have identical variances by default.

    Parameters
    ----------
    x : array
        First input signal.
    y : array
        Second input signal.

    Returns
    -------
    statistic : float
        t-statistic. The t-statistic is used in a t-test to determine
        if you should support or reject the null hypothesis.
    pvalue : float
        Two-sided p-value.

    Raises
    ------
    ValueError
        If the input signals do not have the same length.

    """

    # check inputs
    if x is None:
        raise TypeError("Please specify the first input signal.")

    if y is None:
        raise TypeError("Please specify the second input signal.")

    # ensure numpy
    x = np.array(x)
    y = np.array(y)

    n = len(x)

    if n != len(y):
        raise ValueError("Input signals must have the same length.")

    statistic, pvalue = ttest_ind(x, y)

    return statistic, pvalue


def histogram(signal=None, bins=5, normalize=True):
    """Compute histogram of the input signal.

    Parameters
    ----------
    signal : array
        Input signal.
    bins : int, optional
        Number of histogram bins. Default is 5.
    normalize : bool, optional
        Whether to normalize the histogram counts. Default is True.

    Returns
    -------
    hist{bin}_bins : float
        Number of counts of the bin. If `normalize` is True, the counts are normalized.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure input formats
    signal = np.array(signal)
    bins = int(bins)

    # initialize output
    out = utils.ReturnTuple((), ())

    # compute histogram
    hist = np.histogram(signal, bins=bins)[0]
    if normalize:
        hist = hist / np.sum(hist)  # normalization

    # add counts
    for index, count in enumerate(hist):
        out = out.append(count, 'hist_' + str(index+1) + '_' + str(bins))

    return out


def quartiles(signal=None):
    """Compute quartile features of the signal.

    Parameters
    ----------
    signal : array
        Input signal.

    Returns
    -------
    q1 : float
        First quartile.
    q2 : float
        Second quartile, also known as median.
    q3 : float
        Third quartile.
    iqr : float
        Interquartile range.
    midhinge : float
        Midhinge.
    trimean : float
        Trimean.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure input formats
    signal = np.array(signal)

    # initialize output
    out = utils.ReturnTuple((), ())

    # compute quartiles
    q1, q2, q3 = np.quantile(signal, [0.25, 0.5, 0.75])
    out = out.append([q1, q2, q3], ['q1', 'q2', 'q3'])

    # iqr
    iqr = q3 - q1
    out = out.append(iqr, 'iqr')

    # midhinge
    midhinge = (q3 + q1) / 2
    out = out.append(midhinge, 'midhinge')

    # trimean
    trimean = (q2 + midhinge) / 2
    out = out.append(trimean, 'trimean')

    return out


def diff_stats(signal=None, stats_only=True):
    """Compute statistical features from the first signal differences, second
    signal differences and absolute signal differences.

    Parameters
    ----------
    signal : array
        Input signal.
    stats_only : bool, optional
        Whether to output only statistical features. Default is True.

    Returns
    -------
    {diff} : array
        Difference signal. {diff} can be 'diff', 'diff2' or 'abs_diff'.
    {diff}_mean : float
        Mean of the difference signal.
    {diff}_median : float
        Median of the difference signal.
    {diff}_min : float
        Minimum of the difference signal.
    {diff}_max : float
        Maximum of the difference signal.
    {diff}_max_amp : float
        Maximum amplitude of the difference signal.
    {diff}_range : float
        Range of the difference signal.
    {diff}_var : float
        Variance of the difference signal.
    {diff}_std : float
        Standard deviation of the difference signal.
    {diff}_sum : float
        Sum of the difference signal.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    # initialize output
    out = utils.ReturnTuple((), ())

    # compute differences
    sig_diff = np.diff(signal)
    sig_diff_2 = np.diff(sig_diff)
    sig_diff_abs = np.abs(sig_diff)

    diffs = [sig_diff, sig_diff_2, sig_diff_abs]
    labels = ['firstdiff', 'seconddiff', 'absdiff']

    # extract features
    for diff, label in zip(diffs, labels):
        # add to output
        if not stats_only:
            out = out.append(diff, label)

        # compute stats
        diff_stat = tools.signal_stats(diff)

        # add to output
        for arg, name in zip(diff_stat, diff_stat.keys()):
            out = out.append(arg, label + '_' + name)

        # sum
        sum_ = np.sum(diff)
        out = out.append(sum_, label + '_' + 'sum')

    return out



# -*- coding: utf-8 -*-
"""
biosppy.inter_plotting.ecg
-------------------

This module provides an interactive display option for the ACC plot.

:copyright: (c) 2015-2018 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.

"""


# Imports
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from tkinter import *
import tkinter.font as tkFont
import sys
import os
import matplotlib.backends.backend_tkagg as tkagg

# Globals
from biosppy import utils

MAJOR_LW = 2.5
MINOR_LW = 1.5
MAX_ROWS = 10


def plot_acc(ts=None, raw=None, vm=None, sm=None, spectrum=None, path=None):
    """Create a summary plot from the output of signals.acc.acc.

    Parameters
    ----------
    ts : array
        Signal time axis reference (seconds).
    raw : array
        Raw ACC signal.
    vm : array
        Vector Magnitude feature of the signal.
    sm : array
        Signal Magnitude feature of the signal
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show the plot immediately.

    """

    raw_t = np.transpose(raw)
    acc_x, acc_y, acc_z = raw_t[0], raw_t[1], raw_t[2]

    root = Tk()
    root.resizable(False, False)  # default
    fig, axs_1 = plt.subplots(3, 1)
    axs_1[0].plot(ts, acc_x, linewidth=MINOR_LW, label="Raw acc along X", color="C0")
    axs_1[1].plot(ts, acc_y, linewidth=MINOR_LW, label="Raw acc along Y", color="C1")
    axs_1[2].plot(ts, acc_z, linewidth=MINOR_LW, label="Raw acc along Z", color="C2")

    axs_1[0].set_ylabel("Amplitude ($m/s^2$)")
    axs_1[1].set_ylabel("Amplitude ($m/s^2$)")
    axs_1[2].set_ylabel("Amplitude ($m/s^2$)")
    axs_1[2].set_xlabel("Time (s)")

    fig.suptitle("Acceleration signals")

    global share_axes_check_box

    class feature_figure:
        def __init__(self, reset=False):
            self.figure, self.axes = plt.subplots(1, 1)
            if not reset:
                self.avail_plots = []

        def on_xlims_change(self, event_ax):
            print("updated xlims: ", event_ax.get_xlim())

        def set_labels(self, x_label, y_label):
            self.axes.set_xlabel(x_label)
            self.axes.set_ylabel(y_label)

        def hide_content(self):
            # setting every plot element to white
            self.axes.spines["bottom"].set_color("white")
            self.axes.spines["top"].set_color("white")
            self.axes.spines["right"].set_color("white")
            self.axes.spines["left"].set_color("white")
            self.axes.tick_params(axis="x", colors="white")
            self.axes.tick_params(axis="y", colors="white")

        def show_content(self):
            # setting every plot element to black
            self.axes.spines["bottom"].set_color("black")
            self.axes.spines["top"].set_color("black")
            self.axes.spines["right"].set_color("black")
            self.axes.spines["left"].set_color("black")
            self.axes.tick_params(axis="x", colors="black")
            self.axes.tick_params(axis="y", colors="black")
            self.axes.legend()

        def draw_in_canvas(self, root: Tk, grid_params=None):
            if grid_params is None:
                grid_params = {
                    "row": 2,
                    "column": 1,
                    "columnspan": 4,
                    "sticky": "w",
                    "padx": 10,
                }
            # add an empty canvas for plotting
            self.canvas = matplotlib.backends.backend_tkagg.FigureCanvasTkAgg(self.figure, master=root)
            self.canvas.get_tk_widget().grid(**grid_params)
            self.canvas.draw()
            # self.axes.callbacks.connect('xlim_changed', on_xlims_change)

        def dump_canvas(self, root):
            self.axes.clear()
            self.figure.clear()
            self.figure.canvas.draw_idle()
            self.canvas = matplotlib.backends.backend_tkagg.FigureCanvasTkAgg(self.figure, master=root)
            self.canvas.get_tk_widget().destroy()

        def add_toolbar(self, root: Tk, grid_params=None):
            if grid_params is None:
                grid_params = {"row": 5, "column": 1, "columnspan": 2, "sticky": "w"}
            toolbarFramefeat = Frame(master=root)
            toolbarFramefeat.grid(**grid_params)

            toolbarfeat = matplotlib.backends.backend_tkagg.NavigationToolbar2Tk(self.canvas, toolbarFramefeat)
            toolbarfeat.update()

        def add_plot(self, feature_name: str, xdata, ydata, linewidth, label, color):
            feature_exists = False
            for features in self.avail_plots:
                if features["feature_name"] == feature_name:
                    feature_exists = True
                    break
            if not feature_exists:
                self._add_plot(feature_name, xdata, ydata, linewidth, label, color)

        def _add_plot(self, feature_name: str, xdata, ydata, linewidth, label, color):

            plot_params = {
                "feature_name": feature_name,
                "x": xdata,
                "y": ydata,
                "linewidth": linewidth,
                "label": label,
                "color": color,
            }

            plot_data = dict(plot_params)
            self.avail_plots.append(plot_data)
            del plot_params["feature_name"]
            del plot_params["x"]
            del plot_params["y"]
            self.axes.plot(xdata, ydata, **plot_params)
            self.show_content()

        def remove_plot(self, feature_name: str):
            if self.avail_plots:
                removed_index = [
                    i
                    for i, x in enumerate(self.avail_plots)
                    if x["feature_name"] == feature_name
                ]
                if len(removed_index) == 1:
                    self._remove_plot(removed_index)

        def _remove_plot(self, removed_index):
            del self.avail_plots[removed_index[0]]
            self.__init__(reset=True)

            for params in self.avail_plots:
                temp_params = dict(params)
                del temp_params["feature_name"]
                del temp_params["x"]
                del temp_params["y"]
                self.axes.plot(params["x"], params["y"], **temp_params)

            if self.avail_plots == []:
                self.hide_content()
            else:
                self.show_content()

        def get_axes(self):
            return self.axes

        def get_figure(self):
            return self.figure

    # save to file
    if path is not None:
        path = utils.normpath(path)
        root_, ext = os.path.splitext(path)
        ext = ext.lower()
        if ext not in ["png", "jpg"]:
            path = root_ + ".png"

        fig.savefig(path, dpi=200, bbox_inches="tight")

    # window title
    root.wm_title("BioSPPy: acceleration signal")

    root.columnconfigure(0, weight=4)
    root.columnconfigure(1, weight=1)
    root.columnconfigure(2, weight=1)
    root.columnconfigure(3, weight=1)
    root.columnconfigure(4, weight=1)

    helv = tkFont.Font(family="Helvetica", size=20)

    # checkbox
    show_features_var = IntVar()
    share_axes_var = IntVar()

    def show_features():
        global feat_fig
        global toolbarfeat
        global share_axes_check_box
        if show_features_var.get() == 0:
            drop_features2.get_menu().config(state="disabled")
            domain_feat_btn.config(state="disabled")

            # remove canvas for plotting
            feat_fig.dump_canvas(root)

        if show_features_var.get() == 1:
            # enable option menu for feature selection
            drop_features2.get_menu().config(state="normal")
            domain_feat_btn.config(state="normal")
            # canvas_features.get_tk_widget().grid(row=2, column=1, columnspan=1, sticky='w', padx=10)

            # add an empty canvas for plotting
            feat_fig = feature_figure()
            share_axes_check_box = Checkbutton(
                root,
                text="Share axes",
                variable=share_axes_var,
                onvalue=1,
                offvalue=0,
                command=lambda feat_fig=feat_fig: share_axes(feat_fig.get_axes()),
            )
            share_axes_check_box.config(font=helv)
            share_axes_check_box.grid(row=4, column=1, sticky=W)

            feat_fig.hide_content()
            feat_fig.draw_in_canvas(root)

    def share_axes(ax2):
        if share_axes_var.get() == 1:
            axs_1[0].get_shared_x_axes().join(axs_1[0], ax2)
            axs_1[1].get_shared_x_axes().join(axs_1[1], ax2)
            axs_1[2].get_shared_x_axes().join(axs_1[2], ax2)

        else:
            for ax in axs_1:
                ax.get_shared_x_axes().remove(ax2)
                ax2.get_shared_x_axes().remove(ax)
                ax.autoscale()
                canvas_raw.draw()

    check1 = Checkbutton(
        root,
        text="Show features",
        variable=show_features_var,
        onvalue=1,
        offvalue=0,
        command=show_features,
    )
    check1.config(font=helv)
    check1.grid(row=0, column=0, sticky=W)

    # FEATURES to be chosen
    clicked_features = StringVar()
    clicked_features.set("features")

    def domain_func():
        global share_axes_check_box

        if feat_domain_var.get() == 1:
            domain_feat_btn["text"] = "Domain: frequency"
            feat_domain_var.set(0)
            feat_fig.remove_plot("VM")
            feat_fig.remove_plot("SM")
            feat_fig.draw_in_canvas(root)
            drop_features2.reset()
            drop_features2.reset_fields(["Spectra"])
            share_axes_check_box.config(state="disabled")
            share_axes_var.set(0)

        else:
            domain_feat_btn["text"] = "Domain: time"
            feat_domain_var.set(1)
            feat_fig.remove_plot("SPECTRA X")
            feat_fig.remove_plot("SPECTRA Y")
            feat_fig.remove_plot("SPECTRA Z")
            feat_fig.draw_in_canvas(root)
            drop_features2.reset()
            drop_features2.reset_fields(["VM", "SM"])
            share_axes_check_box.config(state="normal")

    feat_domain_var = IntVar()
    feat_domain_var.set(1)

    class feat_menu:
        def __init__(self, fieldnames: list, entry_name="Select Features", font=helv):
            self.feat_menu = Menubutton(root, text=entry_name, relief="raised")
            self.feat_menu.grid(row=0, column=2, sticky=W)
            self.feat_menu.menu = Menu(self.feat_menu, tearoff=0)
            self.feat_menu["menu"] = self.feat_menu.menu
            self.feat_menu["font"] = font
            self.font = font
            self.feat_activation = {}

            # setting up disabled fields
            for field in fieldnames:
                self.feat_activation[field] = False

            for field in fieldnames:
                self.feat_menu.menu.add_command(
                    label=field,
                    font=helv,
                    command=lambda field=field: self.update_field(field),
                    foreground="gray",
                )
            self.fieldnames = fieldnames

            self.feat_menu.update()

        def reset(self, entry_name="Select Features"):
            self.feat_menu = Menubutton(root, text=entry_name, relief="raised")
            self.feat_menu.grid(row=0, column=2, sticky=W)
            self.feat_menu.menu = Menu(self.feat_menu, tearoff=0)
            self.feat_menu["menu"] = self.feat_menu.menu
            self.feat_menu["font"] = self.font
            self.feat_menu.update()

        def update_field(self, field):

            self.feat_activation[field] = not self.feat_activation[field]
            self.feat_menu.configure(text=field)  # Set menu text to the selected event

            self.reset()

            for field_ in self.fieldnames:
                if self.feat_activation[field_]:
                    self.feat_menu.menu.add_command(
                        label=field_,
                        font=helv,
                        command=lambda field=field_: self.update_field(field),
                    )
                else:
                    self.feat_menu.menu.add_command(
                        label=field_,
                        font=helv,
                        command=lambda field=field_: self.update_field(field),
                        foreground="gray",
                    )

                if field == "SM":
                    if not self.feat_activation[field]:
                        feat_fig.remove_plot("SM")
                        if any(self.feat_activation.values()):
                            feat_fig.set_labels("Time (s)", "Amplitude ($m/s^2$)")

                    else:
                        feat_fig.add_plot(
                            "SM",
                            ts,
                            sm,
                            linewidth=MINOR_LW,
                            label="Signal Magnitude feature",
                            color="C4",
                        )
                        feat_fig.set_labels("Time (s)", "Amplitude ($m/s^2$)")

                    feat_fig.draw_in_canvas(root)
                    feat_fig.add_toolbar(root)

                elif field == "VM":
                    if not self.feat_activation[field]:
                        feat_fig.remove_plot("VM")
                        if any(self.feat_activation.values()):
                            feat_fig.set_labels("Time (s)", "Amplitude ($m/s^2$)")
                    else:
                        feat_fig.add_plot(
                            "VM",
                            ts,
                            vm,
                            linewidth=MINOR_LW,
                            label="Vector Magnitude feature",
                            color="C3",
                        )
                        feat_fig.set_labels("Time (s)", "Amplitude ($m/s^2$)")

                    feat_fig.draw_in_canvas(root)
                    feat_fig.add_toolbar(root)

                elif field == "Spectra":
                    if not self.feat_activation[field]:
                        feat_fig.remove_plot("SPECTRA X")
                        feat_fig.remove_plot("SPECTRA Y")
                        feat_fig.remove_plot("SPECTRA Z")

                    else:

                        feat_fig.add_plot(
                            "SPECTRA X",
                            spectrum["freq"]["x"],
                            spectrum["abs_amp"]["x"],
                            linewidth=MINOR_LW,
                            label="Spectrum along X",
                            color="C0",
                        )
                        feat_fig.draw_in_canvas(root)

                        feat_fig.add_plot(
                            "SPECTRA Y",
                            spectrum["freq"]["y"],
                            spectrum["abs_amp"]["y"],
                            linewidth=MINOR_LW,
                            label="Spectrum along Y",
                            color="C1",
                        )
                        feat_fig.draw_in_canvas(root)

                        feat_fig.add_plot(
                            "SPECTRA Z",
                            spectrum["freq"]["z"],
                            spectrum["abs_amp"]["z"],
                            linewidth=MINOR_LW,
                            label="Spectrum along Z",
                            color="C2",
                        )
                        feat_fig.set_labels(
                            "Frequency ($Hz$)", "Normalized Amplitude [a.u.]"
                        )

                    feat_fig.draw_in_canvas(root)
                    feat_fig.add_toolbar(root)

                self.feat_menu.config(state="normal")
                self.feat_menu.update()

        def reset_fields(self, fieldnames):
            self.feat_activation = {}

            # setting up disabled fields
            for field in fieldnames:
                self.feat_activation[field] = False

            for field in fieldnames:
                self.feat_menu.menu.add_command(
                    label=field,
                    font=helv,
                    command=lambda field=field: self.update_field(field),
                    foreground="gray",
                )

            self.fieldnames = fieldnames

            self.feat_menu.update()

        def get_menu(self):
            return self.feat_menu

    domain_feat_btn = Button(root, text="Domain: time", command=domain_func)
    domain_feat_btn.config(font=helv, state="disabled")
    domain_feat_btn.grid(row=0, column=1, sticky=W, padx=10)

    temp_features = ["VM", "SM"]

    drop_features2 = feat_menu(temp_features, entry_name="Select Features", font=helv)
    drop_features2.get_menu().config(state="disabled")
    drop_features2.get_menu().update()

    canvas_raw = matplotlib.backends.backend_tkagg.FigureCanvasTkAgg(fig, master=root)
    canvas_raw.get_tk_widget().grid(row=2, column=0, columnspan=1, sticky="w", padx=10)
    canvas_raw.draw()

    toolbarFrame = Frame(master=root)
    toolbarFrame.grid(row=5, column=0, columnspan=1, sticky=W)
    toolbar = matplotlib.backends.backend_tkagg.NavigationToolbar2Tk(canvas_raw, toolbarFrame)
    toolbar.update()

    # Add key functionality
    def on_key(event):
        # print('You pressed {}'.format(event.key))
        matplotlib.backend_bases.key_press_handler(event, canvas_raw, toolbar)

    canvas_raw.mpl_connect("key_press_event", on_key)

    # tkinter main loop
    mainloop()



# -*- coding: utf-8 -*-
"""
biosppy.signals.acc
-------------------

This module provides methods to process Acceleration (ACC) signals.
Implemented code assumes ACC acquisition from a 3 orthogonal axis reference system.

:copyright: (c) 2015-2023 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.

Authors:
Afonso Ferreira
Diogo Vieira

"""

# Imports
from __future__ import absolute_import, division, print_function
from six.moves import range
import warnings

# 3rd party
import numpy as np

# local
from .. import plotting, utils
from biosppy.inter_plotting import acc as inter_plotting


def acc(signal=None, sampling_rate=100.0, units=None, path=None, show=True, interactive=False):
    """Process a raw ACC signal and extract relevant signal features using
    default parameters.

    Parameters
    ----------
    signal : array
        Raw ACC signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    units : str, optional
        The units of the input signal. If specified, the plot will have the
        y-axis labeled with the corresponding units.
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show a summary plot.
    interactive : bool, optional
        If True, shows an interactive plot.

    Returns
    -------
    ts : array
        Signal time axis reference (seconds).
    signal : array
        Raw (unfiltered) ACC signal.
    vm : array
        Vector Magnitude feature of the signal.
    sm : array
        Signal Magnitude feature of the signal.
    freq_features : dict
        Positive Frequency domains (Hz) of the signal.
    amp_features : dict
        Normalized Absolute Amplitudes of the signal.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    sampling_rate = float(sampling_rate)

    # extract features
    vm_features, sm_features = time_domain_feature_extractor(signal=signal)
    freq_features, abs_amp_features = frequency_domain_feature_extractor(
        signal=signal, sampling_rate=sampling_rate
    )

    # get time vectors
    length = len(signal)
    T = (length - 1) / sampling_rate
    ts = np.linspace(0, T, length, endpoint=True)

    # plot
    if show:
        if interactive:
            inter_plotting.plot_acc(
                ts=ts,  # plotting.plot_acc
                raw=signal,
                vm=vm_features,
                sm=sm_features,
                spectrum={"freq": freq_features, "abs_amp": abs_amp_features},
                path=path,
            )
        else:
            plotting.plot_acc(
                ts=ts,  # plotting.plot_acc
                raw=signal,
                vm=vm_features,
                sm=sm_features,
                units=units,
                path=path,
                show=True,
            )

    # output
    args = (ts, signal, vm_features, sm_features, freq_features, abs_amp_features)
    names = ("ts", "signal", "vm", "sm", "freq", "abs_amp")

    return utils.ReturnTuple(args, names)


def activity_index(signal=None, sampling_rate=100.0, window_1=5, window_2=60):
    """
    Compute the activity index of an ACC signal.

    Follows the method described in [Lin18]_, the activity index is computed as
    follows:
    
    1) Calculate the ACC magnitude if the signal is triaxial
    2) Calculate the standard deviation of the ACC magnitude for each
    'window_1' seconds
    3) The activity index will be the mean standard deviation for each
    'window_2' seconds
    
    Parameters
    ----------
    signal : array
        Raw ACC signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    window_1 : int, float, optional
        Window length (seconds) for the first moving average filter. Default: 5
    window_2 : int, float, optional
        Window length (seconds) for the second moving average filter. Default: 60

    Returns
    -------
    ts : array
        Time axis reference (seconds).
    activity_index : array
        Activity index of the signal.

    References
    ----------
    .. [Lin18] W.-Y. Lin, V. Verma, M.-Y. Lee, C.-S. Lai, Activity
    Monitoring with a Wrist-Worn, Accelerometer-Based Device, Micromachines 9
    (2018) 450

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")
    
    signal = np.array(signal)
    if signal.shape[1] != 3:
        raise TypeError("Input signal must be triaxial.")

    # window size 1
    window_1 = int(sampling_rate * window_1) 
    # window size 2
    window_2 = int(sampling_rate * window_2) 

    # acc magnitude
    acc_magnitude = np.sum(signal**2, axis=1)**0.5
    # activity index is the sum of 12 standard deviations computed for 5s
    # periods
    if window_2 > len(acc_magnitude):
        warnings.warn("'window_2' must be smaller than the signal length. The"
                      "activity index will be computed for the whole signal.")
        window_2 = len(acc_magnitude)
    # last activity index is in (len(acc_magnitude) - len(acc_magnitude) % window_2)
    acc_magnitude = acc_magnitude[:len(acc_magnitude) - len(acc_magnitude) % window_2]

    # segment magnitude in segments with size window_1
    segments = acc_magnitude.reshape(-1, window_1)

    # compute standard deviation for each segment
    activity_index_ = np.std(segments, axis=1)

    # compute activity index (average standard deviation) for each window_2
    activity_index = np.mean(activity_index_.reshape(-1, int(window_2/window_1)), axis=1)

    # timestamps are the end of each window_2 (remove the first one)
    ts = np.arange(0, len(acc_magnitude)+1, int(window_2))[1:]

    # output
    args = (ts, activity_index)
    names = ("ts", "activity_index")

    return utils.ReturnTuple(args, names)


def time_domain_feature_extractor(signal=None):
    """Extracts the vector magnitude and signal magnitude features from an input ACC signal, given the signal itself.

    Parameters
    ----------
    signal : array
        Input ACC signal.

    Returns
    -------
    vm_features : array
        Extracted Vector Magnitude (VM) feature.
    sm_features : array
        Extracted Signal Magnitude (SM) feature.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # get acceleration features
    vm_features = np.zeros(signal.shape[0])
    sm_features = np.zeros(signal.shape[0])

    for i in range(signal.shape[0]):
        vm_features[i] = np.linalg.norm(
            np.array([signal[i][0], signal[i][1], signal[i][2]])
        )
        sm_features[i] = (abs(signal[i][0]) + abs(signal[i][1]) + abs(signal[i][2])) / 3

    return utils.ReturnTuple((vm_features, sm_features), ("vm", "sm"))


def frequency_domain_feature_extractor(signal=None, sampling_rate=100.0):
    """Extracts the FFT from each ACC sub-signal (x, y, z), given the signal itself.

    Parameters
    ----------
    signal : array
        Input ACC signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).

    Returns
    -------
    freq_features : dict
        Dictionary of positive frequencies (Hz) for all sub-signals.
    amp_features : dict
        Dictionary of Normalized Absolute Amplitudes for all sub-signals.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    freq_features = {}
    amp_features = {}

    # get Normalized FFT for each sub-signal
    for ind, axis in zip(range(signal.shape[1]), ["x", "y", "z"]):
        sub_signal = signal[:, ind]

        n = len(sub_signal)
        k = np.arange(n)
        T = n / sampling_rate
        frq = k / T
        freq_features[axis] = frq[range(n // 2)]

        amp = np.fft.fft(sub_signal) / n
        amp_features[axis] = abs(amp[range(n // 2)])

    return utils.ReturnTuple((freq_features, amp_features), ("freq", "abs_amp"))




# -*- coding: utf-8 -*-
"""
biosppy.inter_plotting.ecg
-------------------

This module provides an interactive display option for the ECG plot.

:copyright: (c) 2015-2018 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.

"""

# Imports
from matplotlib import gridspec
import matplotlib

# from matplotlib.backends.backend_wx import *
import matplotlib.pyplot as plt
import numpy as np
from tkinter import *
import os
from biosppy import utils
import matplotlib.backends.backend_tkagg as tkagg

MAJOR_LW = 2.5
MINOR_LW = 1.5
MAX_ROWS = 10


def plot_ecg(
    ts=None,
    raw=None,
    filtered=None,
    rpeaks=None,
    templates_ts=None,
    templates=None,
    heart_rate_ts=None,
    heart_rate=None,
    path=None,
    show=True,
):
    """Create a summary plot from the output of signals.ecg.ecg.

    Parameters
    ----------
    ts : array
        Signal time axis reference (seconds).
    raw : array
        Raw ECG signal.
    filtered : array
        Filtered ECG signal.
    rpeaks : array
        R-peak location indices.
    templates_ts : array
        Templates time axis reference (seconds).
    templates : array
        Extracted heartbeat templates.
    heart_rate_ts : array
        Heart rate time axis reference (seconds).
    heart_rate : array
        Instantaneous heart rate (bpm).
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show the plot immediately.

    """

    # creating a root widget
    root_tk = Tk()
    root_tk.resizable(False, False)  # default

    fig_raw, axs_raw = plt.subplots(3, 1, sharex=True)
    fig_raw.suptitle("ECG Summary")

    # raw signal plot (1)
    axs_raw[0].plot(ts, raw, linewidth=MAJOR_LW, label="Raw", color="C0")
    axs_raw[0].set_ylabel("Amplitude")
    axs_raw[0].legend()
    axs_raw[0].grid()

    # filtered signal with R-Peaks (2)
    axs_raw[1].plot(ts, filtered, linewidth=MAJOR_LW, label="Filtered", color="C0")

    ymin = np.min(filtered)
    ymax = np.max(filtered)
    alpha = 0.1 * (ymax - ymin)
    ymax += alpha
    ymin -= alpha

    # adding the R-Peaks
    axs_raw[1].vlines(
        ts[rpeaks], ymin, ymax, color="m", linewidth=MINOR_LW, label="R-peaks"
    )

    axs_raw[1].set_ylabel("Amplitude")
    axs_raw[1].legend(loc="upper right")
    axs_raw[1].grid()

    # heart rate (3)
    axs_raw[2].plot(heart_rate_ts, heart_rate, linewidth=MAJOR_LW, label="Heart Rate")
    axs_raw[2].set_xlabel("Time (s)")
    axs_raw[2].set_ylabel("Heart Rate (bpm)")
    axs_raw[2].legend()
    axs_raw[2].grid()

    canvas_raw = matplotlib.backends.backend_tkagg.FigureCanvasTkAgg(fig_raw, master=root_tk)
    canvas_raw.get_tk_widget().grid(
        row=0, column=0, columnspan=1, rowspan=6, sticky="w"
    )
    canvas_raw.draw()

    toolbarFrame = Frame(master=root_tk)
    toolbarFrame.grid(row=6, column=0, columnspan=1, sticky=W)
    toolbar = matplotlib.backends.backend_tkagg.NavigationToolbar2Tk(canvas_raw, toolbarFrame)
    toolbar.update()

    fig = fig_raw

    fig_2 = plt.Figure()
    gs = gridspec.GridSpec(6, 1)

    axs_2 = fig_2.add_subplot(gs[:, 0])

    axs_2.plot(templates_ts, templates.T, "m", linewidth=MINOR_LW, alpha=0.7)
    axs_2.set_xlabel("Time (s)")
    axs_2.set_ylabel("Amplitude")
    axs_2.set_title("Templates")
    axs_2.grid()

    grid_params = {"row": 0, "column": 1, "columnspan": 2, "rowspan": 6, "sticky": "w"}
    canvas_2 = matplotlib.backends.backend_tkagg.FigureCanvasTkAgg(fig_2, master=root_tk)
    canvas_2.get_tk_widget().grid(**grid_params)
    canvas_2.draw()

    toolbarFrame_2 = Frame(master=root_tk)
    toolbarFrame_2.grid(row=6, column=1, columnspan=1, sticky=W)
    toolbar_2 = matplotlib.backends.backend_tkagg.NavigationToolbar2Tk(canvas_2, toolbarFrame_2)
    toolbar_2.update()

    if show:
        # window title
        root_tk.wm_title("BioSPPy: ECG signal")

        # save to file
        if path is not None:
            path = utils.normpath(path)
            root, ext = os.path.splitext(path)
            ext = ext.lower()
            if ext not in ["png", "jpg"]:
                path_block_1 = "{}-summary{}".format(root, ".png")
                path_block_2 = "{}-templates{}".format(root, ".png")
            else:
                path_block_1 = "{}-summary{}".format(root, ext)
                path_block_2 = "{}-templates{}".format(root, ext)

            fig.savefig(path_block_1, dpi=200, bbox_inches="tight")
            fig_2.savefig(path_block_2, dpi=200, bbox_inches="tight")

        mainloop()

    else:
        # close
        plt.close(fig)



# -*- coding: utf-8 -*-
"""
biosppy.features
----------------

This package provides methods to extract common
signal features in:
    * Time domain
    * Frequency domain
    * Time-frequency domain
    * Phase-space domain
    * Cepstral domain

:copyright: (c) 2015-2023 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# compat
from __future__ import absolute_import, division, print_function
# allow lazy loading
from . import frequency, time, time_freq, cepstral, phase_space

